// =============================================================================
//
//  JagdGeschwader_AirCraft.pas
//   Implements AirCraft physics and management
//
// =============================================================================
//
//   Copyright  2005 by Sascha Willems (www.delphigl.de)
//
// =============================================================================
//
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
//
// =============================================================================
//
//   Entry for the 2005 PGD "Dog-Fight" Competition
//
// =============================================================================

unit JagdGeschwader_AirCraft;

interface

uses
 SDL,
 SDLInput,

 SysUtils,
 Dialogs,
 Windows,
 Forms,
 gl3DS,
 Math,
 fmod,
 fmodtypes,
 MMSystem,
 Textures,
 dglOpenGL,
 glMatrixHelper,
 NewtonImport,
 glFrustum,

 JagdGeschwader_Effects,
 JagdGeschwader_SoundSystem,
 JagdGeschwader_Terrain,
 JagdGeschwader_NewtonCallbacks;

type
 // Camera types ===============================================================
 TCameraType = (ctCloseBehind, ctFarBehind, ctAboveCockpit, ctFreeRotate);
 // Controller types ===========================================================
 TControllerType = (ctKeyboard, ctJoystick);
 // TAirCraft ==================================================================
 TAirCraft = class
   NewtonBody     : PNewtonBody;
   Model3DS       : TAll3DSMesh;
   PlaneMesh      : T3DSMesh;
   RotorMesh      : T3DSMesh;
   RotorAngle     : Single;
   NewtonWorld    : PNewtonWorld;
   Throttle       : Single;
   Health         : Single;
   GunHeat        : Single;
   GunCooling     : Single;
   MuzzleTexID    : glUInt;
   Controller     : TControllerType;
   JoystickID     : Integer;
   OutOfLevel     : Boolean;
   UnderWater     : Boolean;
   IsRemote       : Boolean;
   IsClient       : Boolean;
   Pause          : PBoolean;
   Emitter        : TSmokeAndFireEmitter;
   Camera         : TCameraType;
   Lifes          : Integer;
   Speed          : Byte;
   SpeedModifier  : Single;
   FirePower      : Byte;
   LastBonus      : Byte;
   LastBonusTimer : Single;
   // Timers
   HitTimer       : Single;
   LastHit        : Single;
   RespawnTimer   : Single;
   RespawnBlink   : Single;
   DestroyedTimer : Single;
   Destroyed      : Boolean;
   // Movement controls
   RollLeft       : Boolean;
   RollRight      : Boolean;
   PitchFront     : Boolean;
   PitchBack      : Boolean;
   PaddleLeft     : Boolean;
   PaddleRight    : Boolean;
   IsFiring       : Boolean;
   // Position of needed force points (read from 3DS)
   PosRWing       : TVector3f;
   PosLWing       : TVector3f;
   PosPropeller   : TVector3f;
   PosPaddle      : TVector3f;
   PosGun         : array of TVector3f;
   Matrix         : TMatrix4f;
   // Soundchannel
   ChannelFire    : Integer;
   ChannelMotor   : Integer;
   constructor Create(pFile3DS : String; pNewtonWorld : PNewtonWorld; pStartMatrix : TMatrix4f);
   destructor Destroy; override;
   procedure Render;
   procedure Update(pTimeFactor : Single);
   procedure ProcessInput(pTimeFactor : Single);
   procedure Fire;
   procedure Respawn;
   procedure SetHeight(pHeightAboveTerrain : Single = 100);
   function GetPosition : TVector3f;
  end;
 // TAirCraftArray =============================================================
 PAirCraftArray = ^TAirCraftArray;
 TAirCraftArray = array of TAirCraft;

var
 gAirCraftArray : PAirCraftArray;
 SDLJoystick    : PSDL_Joystick;
 ContactInfo    : record
                  Body0 : PNewtonBody;
                  Body1 : PNewtonBody;
                  end;

const
 HitTimerMax   = 5;
 LastHitTime   = 2.5;
 LowHealth     = 30;
 MaxLifes      = 5;

// Forward-declarations for the newton collision callbacks
function CollisionBegin(const material : PNewtonMaterial; const body0 : PNewtonBody; const body1 : PNewtonBody) : Integer; cdecl;
function CollisionAirCraftAirCraft(const material : PNewtonMaterial; const contact : PNewtonContact) : Integer; cdecl;
function CollisionTerrainAirCraft(const material : PNewtonMaterial; const contact : PNewtonContact) : Integer; cdecl;

implementation

// *****************************************************************************
//  Newton callbacks
// *****************************************************************************
//  These must be global and can't be class methods, since you can't easily use
//  delphi's class functions as c-callbacks
// *****************************************************************************

// =============================================================================
//  AirCraftApplyForces
// =============================================================================
//  Applies all forces that act on a plane. Note that this is geared towards
//  an arcarde feel. So don't try to use this for a realistic aircraft simulation.
// =============================================================================
procedure AirCraftApplyForces(const Body : PNewtonBody); cdecl;
var
 Plane       : TAirCraft;
 LocV        : TVector3f;
 TmpV        : TVector3f;
 TmpM        : TMatrix4f;
 GravityV    : TVector3f;
 MaxForce    : Single;
begin
// Retrieve the plane that this body belongs to
if Assigned(NewtonBodyGetUserData(Body)) then
 Plane := TAirCraft(NewtonBodyGetUserData(Body))
else
 exit;

if Plane.IsRemote then
 exit;

MaxForce := (Plane.Speed/5 * 475) * Plane.SpeedModifier;

// We use buoyancy to simulate "lighter" than air vehicles like our plane. If the
// health of the plane is low, we use other values to make controlling it harder
if Plane.Health >= LowHealth then
 begin
 if Plane.UnderWater then
  begin
  GravityV := V3(0, 3, 0);
  NewtonBodyAddBuoyancyForce(Body, 0.9, 0.9, 0.9, @GravityV[0], nil, nil)
  end
 else
  begin
  GravityV := V3(0, 0.5, 0);
  NewtonBodyAddBuoyancyForce(Body, 0.8, 0.6, 0.6, @GravityV[0], nil, nil);
  end
 end
else
 begin
 GravityV := V3(0, 5, 0);
 NewtonBodyAddBuoyancyForce(Body, 0.9, 0.8, 0.8, @GravityV[0], nil, nil)
 end;

// Forces ======================================================================
// Propeller
TmpV := V3(MaxForce * Plane.Throttle, 0, 0);
LocV := Plane.PosPropeller;
NewtonBodyGetMatrix(Body, @TmpM[0,0]);
Matrix_RotateVect(TmpM, TmpV);
Matrix_RotateVect(TmpM, LocV);
TmpV := V3(LocV[0]+TmpV[0], LocV[1]+TmpV[1], LocV[2]+TmpV[2]);
NewtonBodyAddForce(Body, @TmpV[0]);
// Left wing
TmpV := V3(20, 0 , 0);
LocV := Plane.PosLWing;
NewtonBodyGetMatrix(Body, @TmpM[0,0]);
Matrix_RotateVect(TmpM, TmpV);
Matrix_RotateVect(TmpM, LocV);
TmpV := V3(LocV[0]+TmpV[0], LocV[1]+TmpV[1], LocV[2]+TmpV[2]);
NewtonBodyAddForce(Body, @TmpV[0]);
// Right wing
TmpV := V3(20, 0 , 0);
LocV := Plane.PosRWing;
NewtonBodyGetMatrix(Body, @TmpM[0,0]);
Matrix_RotateVect(TmpM, TmpV);
Matrix_RotateVect(TmpM, LocV);
TmpV := V3(LocV[0]+TmpV[0], LocV[1]+TmpV[1], LocV[2]+TmpV[2]);
NewtonBodyAddForce(Body, @TmpV[0]);
// Paddle
TmpV := V3(10, 0 , 0);
LocV := Plane.PosPaddle;
NewtonBodyGetMatrix(Body, @TmpM[0,0]);
Matrix_RotateVect(TmpM, TmpV);
Matrix_RotateVect(TmpM, LocV);
TmpV := V3(LocV[0]+TmpV[0], LocV[1]+TmpV[1], LocV[2]+TmpV[2]);
NewtonBodyAddForce(Body, @TmpV[0]);
// Pitch =======================================================================
if Plane.PitchFront then
 begin
 TmpV := V3(0,0,-20);
 Matrix_RotateVect(TmpM, TmpV);
 NewtonBodyAddTorque(Body, @TmpV[0]);
 end;
if Plane.PitchBack then
 begin
 TmpV := V3(0,0,20);
 Matrix_RotateVect(TmpM, TmpV);
 NewtonBodyAddTorque(Body, @TmpV[0]);
 end;
// Yaw =========================================================================
if Plane.PaddleLeft then
 begin
 TmpV := V3(0,5,0);
 Matrix_RotateVect(TmpM, TmpV);
 NewtonBodyAddTorque(Body, @TmpV[0]);
 end;
if Plane.PaddleRight then
 begin
 TmpV := V3(0,-5,0);
 Matrix_RotateVect(TmpM, TmpV);
 NewtonBodyAddTorque(Body, @TmpV[0]);
 end;
// Roll ========================================================================
if Plane.RollLeft then
 begin
 TmpV := V3(-40,0,0);
 Matrix_RotateVect(TmpM, TmpV);
 NewtonBodyAddTorque(Body, @TmpV[0]);
 end;
if Plane.RollRight then
 begin
 TmpV := V3(40,0,0);
 Matrix_RotateVect(TmpM, TmpV);
 NewtonBodyAddTorque(Body, @TmpV[0]);
 end;
end;

// =============================================================================
//  CollisionBegin
// =============================================================================
function CollisionBegin( const material : PNewtonMaterial; const body0 : PNewtonBody; const body1 : PNewtonBody ) : Integer; cdecl;
begin
ContactInfo.Body0 := Body0;
ContactInfo.Body1 := Body1;
Result            := 1;
end;

// =============================================================================
//  CollisionTerrainAirCraft
// =============================================================================
//  Get's called when an aircraft collides with the terrain.
// =============================================================================
function CollisionTerrainAirCraft( const material : PNewtonMaterial; const contact : PNewtonContact ) : Integer; cdecl;
var
 AirCraft : TAirCraft;
 i        : Integer;
 Velocity : TVector3f;
begin
Result := 1;
// Find out which body is the aircraft
AirCraft := nil;
for i := 0 to High(gAirCraftArray^) do
 if (gAirCraftArray^[i].NewtonBody = ContactInfo.Body0) or (gAirCraftArray^[i].NewtonBody = ContactInfo.Body1) then
  AirCraft := gAirCraftArray^[i];
if not Assigned(AirCraft) then
 exit;
if AirCraft.LastHit <= 0 then
 begin
 if AirCraft.RespawnTimer > 0 then
  exit;
 AirCraft.HitTimer := HitTimerMax;
 AirCraft.LastHit  := LastHitTime;
 // Damage of hit depends of speed of the aircraft
 NewtonBodyGetVelocity(AirCraft.NewtonBody, @Velocity[0]);
 AirCraft.Health   := AirCraft.Health - 5 * VLength(Velocity);
 end;
end;

// =============================================================================
//  CollisionAirCraftAirCraft
// =============================================================================
//  Get's called when an aircraft collides with another aircraft
// =============================================================================
function CollisionAirCraftAirCraft( const material : PNewtonMaterial; const contact : PNewtonContact ) : Integer; cdecl;
var
 AirCraft : array[0..1] of TAirCraft;
 Velocity : TVector3f;
 i        : Integer;
begin
AirCraft[0] := nil;
AirCraft[1] := nil;
Result      := 1;
if Assigned(NewtonBodyGetUserData(ContactInfo.Body0)) then
 AirCraft[0] := TAirCraft(NewtonBodyGetUserData(ContactInfo.Body0));
if Assigned(NewtonBodyGetUserData(ContactInfo.Body1)) then
 AirCraft[1] := TAirCraft(NewtonBodyGetUserData(ContactInfo.Body1));
if (not Assigned(AirCraft[0])) or (not Assigned(AirCraft[1])) then
 exit;
for i := 0 to 1 do
 if AirCraft[i].LastHit <= 0 then
  begin
  if AirCraft[i].RespawnTimer > 0 then
   continue;
  AirCraft[i].HitTimer := HitTimerMax;
  AirCraft[i].LastHit  := LastHitTime;
  // Damage of hit depends of speed of the aircraft
  NewtonBodyGetVelocity(AirCraft[i].NewtonBody, @Velocity[0]);
  AirCraft[i].Health   := AirCraft[i].Health - 10 * VLength(Velocity);
  end;
end;

// =============================================================================
//  Transform3DSMeshToOrigin
// =============================================================================
procedure Transform3DSMeshToOrigin(p3DSMesh : T3DSMesh);
var
 Origin : TVector3f;
 v      : Integer;
begin
with p3DSMesh do
 begin
 Origin := V3(Minimum.x+(Maximum.x-Minimum.x)/2, Minimum.y+(Maximum.y-Minimum.y)/2, Minimum.z+(Maximum.z-Minimum.z)/2);
 for v := 0 to NumVertex-1 do
  begin
  FVertex[v].x := Vertex[v].x-Origin[0];
  FVertex[v].y := Vertex[v].y-Origin[1];
  FVertex[v].z := Vertex[v].z-Origin[2];
  end;
 end;
end;

// *****************************************************************************
//  TAirCraft
// *****************************************************************************

// =============================================================================
//  TAirCraft.Create
// =============================================================================
constructor TAirCraft.Create(pFile3DS : String; pNewtonWorld : PNewtonWorld; pStartMatrix : TMatrix4f);
var
 Collider : PNewtonCollision;
 Ixx      : Single;
 Iyy      : Single;
 Izz      : Single;
 Mass     : Single;
 vMin     : TVector3f;
 vMax     : TVector3f;
 BoxSize  : TVector3f;
 OffsetM  : TMatrix4f;
 i        : Integer;
begin
// Set some default values
ChannelFire   := -1;
ChannelMotor  := -1;
Throttle      := 0.5;
Health        := 100;
GunHeat       := 0;
GunCooling    := 1;
SpeedModifier := 1;
Camera        := ctCloseBehind;
LastHit       := 0;
Lifes         := MaxLifes;
Destroyed     := False;
//
NewtonWorld := pNewtonWorld;
// Load plane model
Model3DS := TAll3DSMesh.Create(nil);
Model3DS.TexturePath := ExtractFilePath(pFile3DS);
Model3DS.LoadFromFile(pFile3DS);
// Find positioning/helper meshes ==============================================
// Try to find plane mesh
if not Assigned(Model3DS.GetMeshByName('plane')) then
 begin
 raise Exception.Create('TAirCraft.Create->No mesh for plane found in "'+pFile3DS+'"!');
 Application.Terminate;
 end
else
 begin
 PlaneMesh := Model3DS.GetMeshByName('plane');
 PlaneMesh.BuildDisplayList;
 end;
// Find force points
// Right wing
if not Assigned(Model3DS.GetMeshByName('rwing')) then
 begin
 raise Exception.Create('TAirCraft.Create->No helper mesh for right wing force position found in "'+pFile3DS+'"!');
 Application.Terminate;
 end
else
 with Model3DS.GetMeshByName('rwing').GetOrigin do
  PosRWing := V3(x, z, y);
// Left wing
if not Assigned(Model3DS.GetMeshByName('lwing')) then
 begin
 raise Exception.Create('TAirCraft.Create->No helper mesh for left wing force position found in "'+pFile3DS+'"!');
 Application.Terminate;
 end
else
 with Model3DS.GetMeshByName('lwing').GetOrigin do
  PosLWing := V3(x, z, y);
// Propeller
if not Assigned(Model3DS.GetMeshByName('prop')) then
 begin
 raise Exception.Create('TAirCraft.Create->No helper mesh for propeller force position found in "'+pFile3DS+'"!');
 Application.Terminate;
 end
else
 with Model3DS.GetMeshByName('prop').GetOrigin do
  PosPropeller := V3(x, z, y);
// Rotormesh
if not Assigned(Model3DS.GetMeshByName('rotor')) then
 begin
 {raise Exception.Create('TAirCraft.Create->No helper mesh for propeller force position found in "'+pFile3DS+'"!');
 Application.Terminate;}
 end
else
 begin
 RotorMesh := Model3DS.GetMeshByName('rotor');
 Transform3DSMeshToOrigin(RotorMesh);
 RotorMesh.BuildDisplayList;
 end;
// Paddle
if not Assigned(Model3DS.GetMeshByName('paddle')) then
 begin
 raise Exception.Create('TAirCraft.Create->No helper mesh for paddle force position found in "'+pFile3DS+'"!');
 Application.Terminate;
 end
else
 with Model3DS.GetMeshByName('paddle').GetOrigin do
  PosPaddle := V3(x, z, y);
// Guns (can be more than one)
for i := 0 to Model3DS.NumMeshes-1 do
 if Pos('gun', LowerCase(Model3DS.Mesh[i].Name)) > 0 then
  begin
  SetLength(PosGun, Length(PosGun)+1);
  with Model3DS.Mesh[i].GetOrigin do
   PosGun[High(PosGun)] := V3(x, z, y);
  end;
if Length(PosGun) = 0 then
 begin
 raise Exception.Create('TAirCraft.Create->Not one single gun position found in "'+pFile3DS+'"!');
 Application.Terminate;
 end;
Emitter    := TSmokeAndFireEmitter.Create(PosPropeller);
// Now create physics body for plane (use a convex collision hull) =============
if Assigned(Model3DS.GetMeshByName('offset')) then
 begin
 Matrix_SetIdentity(OffSetM);
 with Model3DS.GetMeshByName('offset').GetOrigin do
  Matrix_SetTransform(OffSetM, V3(x, z, y));
 Collider   := NewtonCreateConvexHull(NewtonWorld, PlaneMesh.NumVertex, @PlaneMesh.FVertex[0], SizeOf(TVector3f), @OffsetM[0,0]);
 end
else
 Collider   := NewtonCreateConvexHull(NewtonWorld, PlaneMesh.NumVertex, @PlaneMesh.FVertex[0], SizeOf(TVector3f), nil);
NewtonBody := NewtonCreateBody(NewtonWorld, Collider);
// Calculate a close to real moment of inertia
Mass    := 25;
NewtonBodyGetAABB(NewtonBody, @vMin[0], @vMax[0]);
BoxSize := V3(Abs(vMax[0]-vMin[0]), Abs(vMax[1]-vMin[1]), Abs(vMax[2]-vMin[2]));
Ixx := Mass * (BoxSize[1] * BoxSize[1] + BoxSize[2] * BoxSize[2]) / 12;
Iyy := Mass * (BoxSize[0] * BoxSize[0] + BoxSize[2] * BoxSize[2]) / 12;
Izz := Mass * (BoxSize[0] * BoxSize[0] + BoxSize[1] * BoxSize[1]) / 12;
NewtonBodySetMassMatrix(NewtonBody, Mass, Ixx, Iyy, Izz);
NewtonBodySetMatrix(NewtonBody, @pStartMatrix[0,0]);
NewtonBodySetForceAndTorqueCallBack(NewtonBody, AirCraftApplyForces);
NewtonReleaseCollision(NewtonWorld, Collider);
// The plane should never get into rest state
NewtonBodySetAutoFreeze(NewtonBody, 0);
NewtonWorldUnfreezeBody(NewtonWorld, NewtonBody);
// Assign this plane with the newton body
NewtonBodySetUserData(NewtonBody, self);
NewtonBodySetMaterialGroupID(NewtonBody, MatAirCraft);
end;

// =============================================================================
//  TAirCraft.Destroy
// =============================================================================
destructor TAirCraft.Destroy;
begin
Model3DS.Free;
NewtonDestroyBody(NewtonWorld, NewtonBody);
end;

// =============================================================================
//  TAirCraft.Render
// =============================================================================
procedure TAirCraft.Render;
const
 LightPos     : array[0..3] of Single = (0, -10, 0, 1);
 LightAmbient : array[0..3] of Single = (0.5, 0.5, 0.5, 1);
var
 i    : Integer;
 TmpV : TVector3f;
begin
// Don't render (blink) during a short phase after respawn
if RespawnTimer > 0  then
 if RespawnBlink > 0.35 then
  exit;
glPushMatrix;
 glMultMatrixf(@Matrix[0,0]);
 glEnable(GL_LIGHTING);
 glEnable(GL_LIGHT0);
 glLightfv(GL_LIGHT0, GL_POSITION, @LightPos[0]);
 glLightfv(GL_LIGHT0, GL_AMBIENT,  @LightAmbient[0]);
 glEnable(GL_CULL_FACE);
 glCallList(PlaneMesh.DisplayList);
 glPushMatrix;
  with RotorMesh.GetOrigin do
   glTranslatef(x, z, y);
  glRotatef(RotorAngle*10, 1,0,0);
  glCallList(RotorMesh.DisplayList);
 glPopMatrix;
 glDisable(GL_LIGHTING);
 // Muzzle flashes when firing
 glColor3f(1,1,1);
 glEnable(GL_TEXTURE_2D);
 glDepthMask(False);
 glDisable(GL_CULL_FACE);
 glBindTexture(GL_TEXTURE_2D, MuzzleTexID);
 glEnable(GL_BLEND);
 glBlendFunc(GL_ONE, GL_ONE);
 if (IsFiring) and (Random(10) < 6) then
  if Length(PosGun) > 0 then
   for i := 0 to High(PosGun) do
    begin
    glPushMatrix;
     glTranslatef(PosGun[i][0]-0.5, PosGun[i][1], PosGun[i][2]);
     glBegin(GL_QUADS);
      glTexCoord2f(0,0); glVertex3f(0, -0.5, 0);
      glTexCoord2f(0,1); glVertex3f(2, -0.5, 0);
      glTexCoord2f(1,1); glVertex3f(2,  0.5, 0);
      glTexCoord2f(1,0); glVertex3f(0,  0.5, 0);
      glTexCoord2f(0,0); glVertex3f(0, 0, -0.5);
      glTexCoord2f(0,1); glVertex3f(2, 0, -0.5);
      glTexCoord2f(1,1); glVertex3f(2, 0,  0.5);
      glTexCoord2f(1,0); glVertex3f(0, 0,  0.5);
     glEnd;
    glPopMatrix;
    end;
 glDepthMask(True);
 glDisable(GL_BLEND);
 glEnable(GL_CULL_FACE);
glPopMatrix;
TmpV := Matrix_TansformVector(Matrix, PosPropeller);
Emitter.Position := TmpV;
if Health < LowHealth then
 Emitter.Render;
end;

// =============================================================================
//  TAirCraft.Update
// =============================================================================
procedure TAirCraft.Update(pTimeFactor : Single);
var
 TmpPos : TVector3f;
begin
// Motor sound
if not SoundSystem.Disabled then
 begin
 if ChannelMotor < 0 then
  ChannelMotor := FSOUND_PlaySound(FSOUND_FREE, SoundSystem.GetSample('motor'))
 else
  if not FSOUND_IsPlaying(ChannelMotor) then
   if Health > LowHealth then
    FSOUND_PlaySound(ChannelMotor, SoundSystem.GetSample('motor'))
   else
    FSOUND_PlaySound(ChannelMotor, SoundSystem.GetSample('motorend'));
 if Health > LowHealth then
  FSOUND_SetFrequency(ChannelMotor, Round(6000 + Throttle*4000))
 else
  FSOUND_SetFrequency(ChannelMotor, 12000);
 FSOUND_SetVolumeAbsolute(ChannelMotor, 128);
 end;
// Update matrix
//if not IsRemote then
 NewtonBodyGetMatrix(NewtonBody, @Matrix[0,0]);
// Gun heat
if not Pause^ then
 GunHeat := GunHeat - pTimeFactor * 0.25 * GunCooling;
if GunHeat < 0 then
 GunHeat := 0;
// Timers
if not Pause^ then
 HitTimer := HitTimer - pTimeFactor * 0.75;
if HitTimer < 0 then
 HitTimer := 0;
if not Pause^ then
 LastHit  := LastHit - pTimeFactor * 0.75;
if LastHit < 0 then
 LastHit := 0;
if not Pause^ then
 LastBonusTimer  := LastBonusTimer - pTimeFactor * 0.15;
if LastBonusTimer < 0 then
 LastBonusTimer := 0;
if not Pause^ then
 RespawnTimer  := RespawnTimer - pTimeFactor * 0.05;
if RespawnTimer < 0 then
 RespawnTimer := 0;
if RespawnTimer > 0 then
 begin
 if RespawnBlink < 0 then
  RespawnBlink := 0.7;
 if not Pause^ then
  RespawnBlink  := RespawnBlink - pTimeFactor * 0.05;
 end;
// Out of playground borders
OutOfLevel := False;
TmpPos := GetPosition;
if (TmpPos[0] < -Terrain.Width/2) or (TmpPos[0] > Terrain.Width/2) or
   (TmpPos[2] < -Terrain.Height/2) or (TmpPos[2] > Terrain.Height/2) or
   (TmpPos[1] > 1000) then
    begin
    if not Pause^ then
     Health := Health - 0.15 * pTimeFactor;
    OutOfLevel := True;
    end;
if not Pause^ then
 Emitter.Update(pTimeFactor*0.01);
if not Pause^ then
 RotorAngle := RotorAngle + Throttle * pTimeFactor;
if RotorAngle > 360 then
 RotorAngle := RotorAngle - 360;
// Destroyed?
if (Health <= 0) and (not Pause^) then
 begin
 if Destroyed then
  begin
  DestroyedTimer := DestroyedTimer - pTimeFactor * 0.4;
  if Random(10) = 5 then
   begin
   ExplosionContainer.LastSpawn := 0;
   ExplosionContainer.Spawn(V3(TmpPos[0]-Random+Random, TmpPos[1]-Random+Random, TmpPos[2]-Random+Random), 1+2*Random);
   end;
  if DestroyedTimer < 0 then
   begin
   Respawn;
   Lifes := Lifes-1;
   end;
  end
 else
  begin
  DestroyedTimer := 175; // initially 100
  Destroyed      := True;
  SoundSystem.PlaySample('explosion', False);
  end;
 end;
end;

// =============================================================================
//  TAirCraft.ProcessInput
// =============================================================================
procedure TAirCraft.ProcessInput(pTimeFactor : Single);
var
 Keys : array[0..255] of Boolean;
 i    : Integer;
begin
if Pause^ then
 exit;
if (not IsRemote) or (IsClient) then
 begin
 // Keyboard
 if Controller = ctKeyBoard then
  begin
  for i := 0 to High(Keys) do
   Keys[i] := (GetAsyncKeyState(i) <> 0);
  RollLeft    := Keys[vk_Left];
  RollRight   := Keys[vk_Right];
  PitchFront  := Keys[vk_Up];
  PitchBack   := Keys[vk_Down];
  PaddleLeft  := Keys[Ord('Q')];
  PaddleRight := Keys[Ord('W')];
  IsFiring    := Keys[vk_Space];
  if Keys[vk_add] then
   Throttle := Throttle + 0.02 * pTimeFactor;
  if Keys[vk_subtract] then
   Throttle := Throttle - 0.02 * pTimeFactor;
  end
 // Joystick/Gamepad
 else
  begin
  // SDL
  SDL_JoystickUpdate;
  Throttle    := 1-((SDL_JoystickGetAxis(SDLJoystick, 2)+32768)/65535);
  RollLeft    := (SDL_JoystickGetAxis(SDLJoystick, 0) < -8192);
  RollRight   := (SDL_JoystickGetAxis(SDLJoystick, 0) >  8192);
  PitchFront  := (SDL_JoystickGetAxis(SDLJoystick, 1) < -8192);
  PitchBack   := (SDL_JoystickGetAxis(SDLJoystick, 1) >  8192);
  PaddleLeft  := (SDL_JoystickGetAxis(SDLJoystick, 3) < -8192);
  PaddleRight := (SDL_JoystickGetAxis(SDLJoystick, 3) >  8192);
  IsFiring    := (SDL_JoystickGetButton(SDLJoystick, 0) = 1);
  end;
 end;
// Check and clamp some values
if IsFiring then
 Fire;
if GunHeat > 120 then
 IsFiring := False;
if Throttle < 0 then
 Throttle := 0;
if Throttle > 1 then
 Throttle := 1;
end;

// =============================================================================
//  TAirCraft.Fire
// =============================================================================
procedure TAirCraft.Respawn;
const
 ZeroV  : array[0..2] of Single = (0,0,0);
var
 NewMat : TMatrix4f;
 NewPos : TVector3f;
begin
ChannelFire    := -1;
ChannelMotor   := -1;
Throttle       := 0.5;
Health         := 100;
GunHeat        := 0;
GunCooling     := 1;
SpeedModifier  := 1;
Camera         := ctCloseBehind;
LastHit        := 0;
RespawnTimer   := 5;
RespawnBlink   := 0.7;
Destroyed      := False;
LastBonusTimer := 0;
NewtonBodySetVelocity(NewtonBody, @ZeroV[0]);
NewtonBodySetOmega(NewtonBody, @ZeroV[0]);
NewtonBodySetTorque(NewtonBody, @ZeroV[0]);
Matrix_SetIdentity(NewMat);
NewPos   := V3((Random-Random) * Terrain.Width/2 * 0.75, 600, (Random-Random) * Terrain.Height/2 * 0.75);
NewPos[1] := GetHeight(NewtonWorld, NewPos)+200;
Matrix_SetTransform(NewMat, NewPos);
NewtonBodySetMatrix(NewtonBody, @NewMat[0,0]);
end;

// =============================================================================
//  TAirCraft.Fire
// =============================================================================
//  Uses newton's raycast interface to detect if another plane has been hit
//  by machine gun fire of this plane. See newton documentation for more info
//  on the raycast interface
// =============================================================================
procedure TAirCraft.Fire;
var
 LineStart : TVector3f;
 LineEnd   : TVector3f;
 HitPos    : TVector3f;
 i,j       : Integer;
begin
// Exit if guns are too hot
if GunHeat < 140 then
 GunHeat := GunHeat + Random*2;
if GunHeat > 110 then
 exit;
// Playback firing sound
if ChannelFire < 0 then
 ChannelFire := FSOUND_PlaySound(FSOUND_FREE, SoundSystem.GetSample('mgun1'))
else
 if not FSOUND_IsPlaying(ChannelFire) then
  FSOUND_PlaySound(ChannelFire, SoundSystem.GetSample('mgun1'));
// Go through all machine guns of the plane
if Length(PosGun) > 0 then
 for i := 0 to High(PosGun) do
  begin
  // Raycasting is done in global space, so we need to get the gun ray from local coords to global coords
  LineStart    := Matrix_TansformVector(Matrix, V3(PosGun[i][0]+225, PosGun[i][1]+Random(5)-Random(5), PosGun[i][2]+Random(5)-Random(5)));
  LineEnd      := Matrix_TansformVector(Matrix, PosGun[i]);
  PickedBody   := nil;
  PickedParam  := 1.1;
  PickedIgnore := NewtonBody;
  NewtonWorldRayCast(NewtonWorld, @LineStart[0], @LineEnd[0], @RayCastFilter, nil);
  // If ray hit another body, check which plane it was
  if Assigned(PickedBody) then
   begin
   for j := 0 to High(gAirCraftArray^) do
    if gAirCraftArray^[j].NewtonBody = PickedBody then
     if (gAirCraftArray^[j].LastHit <= 0) and (gAirCraftArray^[j].RespawnTimer <= 0) then
      begin
      gAirCraftArray^[j].Health   := gAirCraftArray^[j].Health - 5 * FirePower/5;
      gAirCraftArray^[j].HitTimer := HitTimerMax;
      end;
   HitPos := VAdd(LineStart, VScale(VSub(LineEnd, LineStart), PickedParam));
   if PickedBody = Terrain.NewtonBody then
    ExplosionContainer.Spawn(HitPos, 1.2+Random-Random)
   else
    ExplosionContainer.Spawn(VScale(HitPos, 0.98), 2+Random-Random);
   end;
  end;
end;

// =============================================================================
//  TAirCraft.GetPosition
// =============================================================================
function TAirCraft.GetPosition : TVector3f;
begin
Result := V3(Matrix[3,0], Matrix[3,1], Matrix[3,2]);
end;

// =============================================================================
//  TAirCraft.SetHeight
// =============================================================================
procedure TAirCraft.SetHeight(pHeightAboveTerrain : Single = 100);
var
 LineStart : TVector3f;
 LineEnd   : TVector3f;
 HitPos    : TVector3f;
 TmpM      : TMatrix4f;
begin
LineStart    := GetPosition;
LineStart[1] := LineStart[1]-2000;
LineEnd      := GetPosition;
PickedBody   := nil;
PickedParam  := 1.1;
PickedIgnore := NewtonBody;
NewtonWorldRayCast(NewtonWorld, @LineStart[0], @LineEnd[0], @RayCastFilter, nil);
HitPos    := VAdd(LineStart, VScale(VSub(LineEnd, LineStart), PickedParam));
HitPos[1] := HitPos[1] + pHeightAboveTerrain;
Matrix_SetIdentity(TmpM);
Matrix_SetTransform(TmpM, HitPos);
NewtonBodySetMatrix(NewtonBody, @TmpM[0,0]);
end;

end.
