// =============================================================================
//
//  JagdGeschwader_Effects.pas
//
// =============================================================================
//
//   Copyright  2005 by Sascha Willems (www.delphigl.de)
//
// =============================================================================
//
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
//
// =============================================================================
//
//   Entry for the 2005 PGD "Dog-Fight" Competition
//
// =============================================================================

unit JagdGeschwader_Effects;

interface

uses
 Windows,
 SysUtils,

 Textures,
 dglOpenGL,
 glFrustum,
 glMatrixHelper,
 NewtonImport;

type
 // TParticle ==================================================================
 TParticle = record
   Position  : TVector3f;
   Velocity  : TVector3f;
   Life      : Single;
   MaxLife   : Single;
   Texture   : glUInt;
  end;
 // TParticleEmitter ===========================================================
 TParticleEmitter = class
   Particle : array of TParticle;
   Texture  : array of glUInt;
   Position : TVector3f;
   procedure Render; virtual; abstract;
   procedure Update(pTimeFactor : Single); virtual; abstract;
   procedure SpawnParticles(pCount : Integer); virtual; abstract;
   constructor Create(pPosition : TVector3f); virtual; abstract;
  end;
 // TFireEmitter ===============================================================
 TSmokeAndFireEmitter = class(TParticleEmitter)
   procedure Render; override;
   procedure Update(pTimeFactor : Single); override;
   procedure SpawnParticles(pCount : Integer); override;
   constructor Create(pPosition : TVector3f); override;
  end;

 // TExplosion =================================================================
 TExplosion = record
   Position : TVector3f;
   Frame    : Single;
   Size     : Single;
   Remove   : Boolean;
  end;
 // TExplosionContainer ========================================================
 TExplosionContainer = class
   Explosion : array of TExplosion;
   LastSpawn : Single;
   destructor Destroy; override;
   procedure Clear;
   procedure Render;
   procedure Update(pTimeFactor : Single);
   procedure Spawn(pPosition : TVector3f; pSize : Single);
  end;

 // TLensFlare =================================================================
 TLensFlare = class
   Halo      : glUInt;
   BigGlow   : glUInt;
   Glow      : glUInt;
   Streak    : glUInt;
   FadeTimer : Single;
   procedure SetTextures(pHaloID, pBigGlowID, pGlowID, pStreakID : glUInt);
   procedure DrawQuad(pTexture : glUInt; pR, pG, pB, pAlpha : Single; pPosition :  TVector3f; pScale : Single);
   procedure Render(pSunPosition, pCameraPosition : TVector3f; pAlphaFactor : Single; pTimeFactor : Single);
  end;

var
 ExplosionContainer : TExplosionContainer;
 LensFlare          : TLensFlare;

implementation

// *****************************************************************************
//  TSmokeAndFireEmitter
// *****************************************************************************

// =============================================================================
//  TSmokeAndFireEmitter.Create
// =============================================================================
constructor TSmokeAndFireEmitter.Create(pPosition : TVector3f);
begin
Position := pPosition;
end;

// =============================================================================
//  TSmokeAndFireEmitter.SpawnParticles
// =============================================================================
procedure TSmokeAndFireEmitter.SpawnParticles(pCount : Integer);
begin
SetLength(Particle, Length(Particle)+1);
with Particle[High(Particle)] do
 begin
 Position  := V3(Position[0]-Random*0.1+Random*0.1, Position[1]-Random*0.2+Random*0.2, Position[2]-Random*0.1+Random*0.1);
 MaxLife   := 4;
 Life      := MaxLife;
 Velocity  := V3(0, 0.75, 0);
 Texture   := self.Texture[Random(Length(self.Texture))];
 end;
end;

// =============================================================================
//  TSmokeAndFireEmitter.Update
// =============================================================================
procedure TSmokeAndFireEmitter.Update(pTimeFactor : Single);
var
 i : Integer;
begin
if Length(Particle) < 200 then
 if Random(4) = 0 then
  SpawnParticles(1);
if Length(Particle) > 0 then
 for i := 0 to High(Particle) do
  with Particle[i] do
   begin
   Life := Life - pTimeFactor;
   if Life <= 0 then
    begin
    Position := V3(self.Position[0]-Random*2+Random*2, self.Position[1], self.Position[2]-Random*2+Random*2);
    MaxLife  := 4;
    Life     := MaxLife;
    end
   else
    Position := V3(Position[0] + Velocity[0] * pTimeFactor, Position[1] + Velocity[1] * pTimeFactor, Position[2] + Velocity[2] * pTimeFactor);
  end;
end;

// =============================================================================
//  TSmokeAndFireEmitter.Render
// =============================================================================
procedure TSmokeAndFireEmitter.Render;
var
 i       : Integer;
 TmpM    : TMatrix4f;
 TmpV    : TVector3f;
 CurrTex : glUInt;
begin
glBlendFunc(GL_SRC_ALPHA, GL_ONE);
glDepthMask(False);
glDisable(GL_CULL_FACE);
glDisable(GL_LIGHTING);
glEnable(GL_BLEND);
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
glEnable(GL_TEXTURE_2D);
glGetFloatv(GL_MODELVIEW_MATRIX, @TmpM[0,0]);
CurrTex := 0;
glPushMatrix;
 glLoadIdentity;
 if Length(Particle) > 0 then
  for i := 0 to High(Particle) do
   with Particle[i] do
    begin
    if Texture <> CurrTex then
     begin
     glBindTexture(GL_TEXTURE_2D, Texture);
     CurrTex := Texture;
     end;
    TmpV := Matrix_TansformVector(TmpM, Position);
    glColor4f(1,1,1, 0.5*Life/MaxLife);
    glBegin(GL_QUADS);
     glTexCoord2f(1,1); glVertex3f(TmpV[0]-0.8, TmpV[1]-0.8, TmpV[2]);
     glTexCoord2f(1,0); glVertex3f(TmpV[0]-0.8, TmpV[1]+0.8, TmpV[2]);
     glTexCoord2f(0,0); glVertex3f(TmpV[0]+0.8, TmpV[1]+0.8, TmpV[2]);
     glTexCoord2f(0,1); glVertex3f(TmpV[0]+0.8, TmpV[1]-0.8, TmpV[2]);
    glEnd;
    end;
glPopMatrix;
glDepthMask(True);
glDisable(GL_BLEND);
end;

// *****************************************************************************
// TExplosionContainer
// *****************************************************************************

// =============================================================================
// TExplosionContainer.Destroy
// =============================================================================
destructor TExplosionContainer.Destroy;
begin
Clear;
end;

// =============================================================================
// TExplosionContainer.Clear
// =============================================================================
procedure TExplosionContainer.Clear;
begin
SetLength(Explosion, 0);
end;

// =============================================================================
// TExplosionContainer.Render
// =============================================================================
procedure TExplosionContainer.Render;
const
 TexCoord : array[0..15, 0..3] of Single =
   ((0, 0,    0.25, 0.25),  (0.25, 0,     0.5, 0.25),  (0.5, 0,    0.75, 0.25),  (0.75, 0,    1, 0.25),
    (0, 0.25, 0.25, 0.5),   (0.25, 0.25,  0.5, 0.5),   (0.5, 0.25, 0.75, 0.5),   (0.75, 0.25, 1, 0.5),
    (0, 0.5,  0.25, 0.75),  (0.25, 0.5,   0.5, 0.75),  (0.5, 0.5,  0.75, 0.75),  (0.75, 0.5,  1, 0.75),
    (0, 0.75, 0.25, 1),     (0.25, 0.75,  0.5, 1),     (0.5, 0.75, 0.75, 1),     (0.75, 0.75, 1, 1));
var
 i       : Integer;
 TmpM    : TMatrix4f;
 TmpV    : TVector3f;
begin
if Length(Explosion) = 0 then
 exit;
glBlendFunc(GL_SRC_ALPHA, GL_ONE);
glDepthMask(False);
glDisable(GL_CULL_FACE);
glDisable(GL_LIGHTING);
glEnable(GL_BLEND);
glBlendFunc(GL_ONE, GL_ONE);
glEnable(GL_TEXTURE_2D);
glGetFloatv(GL_MODELVIEW_MATRIX, @TmpM[0,0]);
glPushMatrix;
 glLoadIdentity;
 for i := 0 to High(Explosion) do
  with Explosion[i] do
   if not Remove then
    begin
    TmpV := Matrix_TansformVector(TmpM, Position);
    glBegin(GL_QUADS);
     glTexCoord2f(TexCoord[Round(Frame)][0], -TexCoord[Round(Frame)][3]); glVertex3f(TmpV[0]-1*Size, TmpV[1]-1*Size, TmpV[2]);
     glTexCoord2f(TexCoord[Round(Frame)][2], -TexCoord[Round(Frame)][3]); glVertex3f(TmpV[0]+1*Size, TmpV[1]-1*Size, TmpV[2]);
     glTexCoord2f(TexCoord[Round(Frame)][2], -TexCoord[Round(Frame)][1]); glVertex3f(TmpV[0]+1*Size, TmpV[1]+1*Size, TmpV[2]);
     glTexCoord2f(TexCoord[Round(Frame)][0], -TexCoord[Round(Frame)][1]); glVertex3f(TmpV[0]-1*Size, TmpV[1]+1*Size, TmpV[2]);
    glEnd;
    end;
glPopMatrix;
glDepthMask(True);
glDisable(GL_BLEND);
end;

// =============================================================================
// TExplosionContainer.Update
// =============================================================================
procedure TExplosionContainer.Update(pTimeFactor : Single);
var
 i     : Integer;
 TmpEx : TExplosion;
begin
if LastSpawn > 0 then
 LastSpawn := LastSpawn - pTimeFactor * 0.075;
if Length(Explosion) = 0 then
 exit;
for i := 0 to High(Explosion) do
 with Explosion[i] do
  begin
  Frame := Frame + 0.075 * pTimeFactor;
  if Frame > 15 then
   begin
   Frame  := 15;
   Remove := True;
   end;
  end;
// Remove finished explosion
for i := 0 to High(Explosion) do
 if i < High(Explosion) then
  if Explosion[i].Remove then
   begin
   TmpEx                      := Explosion[High(Explosion)];
   Explosion[High(Explosion)] := Explosion[i];
   Explosion[i]               := TmpEx;
   SetLength(Explosion, Length(Explosion)-1);
   end;
end;

// =============================================================================
// TExplosionContainer.Spawn
// =============================================================================
procedure TExplosionContainer.Spawn(pPosition : TVector3f; pSize : Single);
begin
if LastSpawn > 0 then
 exit;
SetLength(Explosion, Length(Explosion)+1);
with Explosion[High(Explosion)] do
 begin
 Size     := pSize;
 Frame    := 0;
 Remove   := False;
 Position := pPosition;
 end;
LastSpawn := 3.5;
end;

// *****************************************************************************
// TLensFlare
// *****************************************************************************

// =============================================================================
// TLensFlare.SetTextures
// =============================================================================
procedure TLensFlare.SetTextures(pHaloID, pBigGlowID, pGlowID, pStreakID : glUInt);
begin
Halo    := pHaloID;
BigGlow := pBigGlowID;
Glow    := pGlowID;
Streak  := pStreakID;
end;

// =============================================================================
// TLensFlare.DrawQuad
// =============================================================================
procedure TLensFlare.DrawQuad(pTexture : glUInt; pR, pG, pB, pAlpha : Single; pPosition :  TVector3f; pScale : Single);
var
 TmpM : TMatrix4f;
 TmpV : TVector3f;
begin
pScale := pScale * 5;
glPushMatrix;
 glGetFloatv(GL_MODELVIEW_MATRIX, @TmpM[0,0]);
 glLoadIdentity;
 TmpV := Matrix_TansformVector(TmpM, pPosition);
 glBindTexture(GL_TEXTURE_2D, pTexture);
 glColor4f(pR, pG, pB, pAlpha);
 glBegin(GL_QUADS);
  glTexCoord2f(0,0); glVertex3f(TmpV[0]-pScale,  TmpV[1]-pScale, TmpV[2]);
  glTexCoord2f(0,1); glVertex3f(TmpV[0]-pScale,  TmpV[1]+pScale, TmpV[2]);
  glTexCoord2f(1,1); glVertex3f(TmpV[0]+pScale,  TmpV[1]+pScale, TmpV[2]);
  glTexCoord2f(1,0); glVertex3f(TmpV[0]+pScale,  TmpV[1]-pScale, TmpV[2]);
 glEnd;
 glColor3f(1,1,1);
glPopMatrix;
end;

// =============================================================================
// TLensFlare.Render
// =============================================================================
procedure TLensFlare.Render(pSunPosition, pCameraPosition : TVector3f; pAlphaFactor : Single; pTimeFactor : Single);
var
 Matrix            : array[0..15] of glFloat;
 m_DirectionVector : TVector3f;
 vLightToCamera    : TVector3f;
 vLightToIntersect : TVector3f;
 LengthV           : Single;
 LengthLtC         : Single;
 Alpha             : Single;
 ptIntersect       : TVector3f;
 ProjMat           : TGLMatrixd4;
 ModelMat          : TGLMatrixd4;
 ViewPort          : TVector4i;
 ZPos              : Single;
 WinCoords         : array[0..2] of Double;
 Tmp               : array[0..1] of Double;
begin
Frustum.Calculate;
if Frustum.IsSphereWithin(pSunPosition[0], pSunPosition[1], pSunPosition[2], 1) then
 begin
 glGetIntegerv(GL_VIEWPORT, @Viewport);
 glGetDoublev(GL_MODELVIEW_MATRIX, @ModelMat);
 glGetDoublev(GL_PROJECTION_MATRIX, @ProjMat);
 gluProject(pSunPosition[0], pSunPosition[1], pSunPosition[2], ModelMat, ProjMat, ViewPort, @WinCoords[0], @WinCoords[1], @WinCoords[2]);
 glReadPixels(Round(WinCoords[0]), Round(WinCoords[1]), 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, @ZPos);
 if ZPos > WinCoords[2] then
  begin
  FadeTimer := FadeTimer + pTimeFactor * 0.25;
  if FadeTimer > 20 then
   FadeTimer := 20;
  end
 else
  if FadeTimer > 0 then
   FadeTimer := FadeTimer - pTimeFactor * 0.15;
 if FadeTimer > 0 then
  begin
  glGetFloatv(GL_MODELVIEW_MATRIX, @Matrix);
  m_DirectionVector := V3(Matrix[8], Matrix[9], -Matrix[10]);
  vLightToCamera    := VSub(pCameraPosition, pSunPosition);
  LengthV           := VLength(vLightToCamera);
  ptIntersect       := VScale(m_DirectionVector, V3(LengthV, LengthV, LengthV));
  ptIntersect       := VAdd(ptIntersect, pCameraPosition);
  vLightToIntersect := VSub(ptIntersect, pSunPosition);
  LengthV           := VLength(vLightToIntersect) * 0.15;
  VNormalize(vLightToIntersect);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  glDisable(GL_DEPTH_TEST);
  glEnable(GL_TEXTURE_2D);
  pAlphaFactor := pAlphaFactor * FadeTimer/20;
  Tmp[0] := ViewPort[2]/2 - WinCoords[0];
  Tmp[1] := ViewPort[3]/2 - (WinCoords[1] / pAlphaFactor);
  LengthLtC :=  Sqrt(Tmp[0]*Tmp[0] + Tmp[1]*Tmp[1]);
  Alpha := ((ViewPort[3]/LengthLtC)/5) * pAlphaFactor;
  DrawQuad(BigGlow, 0.6, 0.6, 0.8, 1*Alpha, pSunPosition, 16);
  DrawQuad(Streak,  0.6, 0.6, 0.8, 1*Alpha, pSunPosition, 16);
  DrawQuad(Glow,    0.8, 0.8, 1, 0.5*Alpha, pSunPosition, 3.5);
  DrawQuad(Glow, 0.9, 0.6, 0.4, 0.5*Alpha, VAdd(VScale(vLightToIntersect, LengthV*0.1), pSunPosition), 0.6);
  DrawQuad(Halo, 0.8, 0.5, 0.6, 0.5*Alpha, VAdd(VScale(vLightToIntersect,LengthV*0.15), pSunPosition), 1.7);
  DrawQuad(Halo, 0.9, 0.2, 0.1, 0.5*Alpha, VAdd(VScale(vLightToIntersect, LengthV*0.175), pSunPosition), 0.83);
  DrawQuad(Halo, 0.7, 0.7, 0.4, 0.5*Alpha, VAdd(VScale(vLightToIntersect, LengthV*0.285), pSunPosition), 1.6);
  DrawQuad(Glow, 0.9, 0.9, 0.2, 0.5*Alpha, VAdd(VScale(vLightToIntersect, LengthV*0.2755), pSunPosition), 0.8);
  DrawQuad(Glow, 0.93, 0.82, 0.73, 0.5*Alpha, VAdd(VScale(vLightToIntersect, LengthV*0.4775), pSunPosition), 1);
  DrawQuad(Halo, 0.7, 0.6, 0.5, 0.5*Alpha, VAdd(VScale(vLightToIntersect, LengthV*0.49), pSunPosition), 1.4);
  DrawQuad(Glow, 0.7, 0.8, 0.3, 0.5*Alpha, VAdd(VScale(vLightToIntersect, LengthV*0.65), pSunPosition), 1.8);
  DrawQuad(Glow, 0.4, 0.3, 0.2, 0.5*Alpha, VAdd(VScale(vLightToIntersect, LengthV*0.63), pSunPosition), 1.4);
  DrawQuad(Halo, 0.7, 0.5, 0.5, 0.5*Alpha, VAdd(VScale(vLightToIntersect, LengthV*0.8), pSunPosition), 1.4);
  DrawQuad(Glow, 0.8, 0.5, 0.1, 0.5*Alpha, VAdd(VScale(vLightToIntersect, LengthV*0.7825), pSunPosition), 0.6);
  DrawQuad(Halo, 0.5, 0.5, 0.7, 0.5*Alpha, VAdd(VScale(vLightToIntersect, LengthV*1), pSunPosition), 1.7);
  DrawQuad(Glow, 0.4, 0.1, 0.9, 0.5*Alpha, VAdd(VScale(vLightToIntersect, LengthV*0.975), pSunPosition), 2);
  glEnable(GL_DEPTH_TEST);
  glDisable(GL_BLEND);
  end
 end;
end;

end.
