// =============================================================================
//
//  JagdGeschwader_GameClass.pas
//
// =============================================================================
//
//   Copyright  2005 by Sascha Willems (www.delphigl.de)
//
// =============================================================================
//
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
//
// =============================================================================
//
//   Entry for the 2005 PGD "Dog-Fight" Competition
//
// =============================================================================

unit JagdGeschwader_GameClass;

interface

uses
 SDL,

 Windows,
 SysUtils,
 MMSystem,
 Classes,
 Forms,
 Math,
 IniFiles,
 Dialogs,

 glOctree,
 dglOpenGL,
 Textures,
 glFont,
 NewtonImport,
 glMatrixHelper,
 glFrustum,
 gl3DS,

 JagdGeschwader_Network,
 JagdGeschwader_Effects,
 JagdGeschwader_SoundSystem,
 JagdGeschwader_NewtonCallbacks,
 JagdGeschwader_Terrain,
 JagdGeschwader_AirCraft;

type
 TSPoint = record
   X,Y : Single;
  end;
 // TMenuItem ===================================================================
 TMenuItem = object
   ActiveTexture : String;
   Action        : String;
   Position      : TSPoint;
   Size          : TSPoint;
  end;
 // TSubMenu ===================================================================
 TSubMenu = class
   MenuItem          : array of TMenuItem;
   BackGroundTexture : String;
   Name              : String;
  end;
 // TGameMenu ==================================================================
 TGameMenu = class(TComponent)
   private
    TextureFont       : TTexFont;
    FVisible          : Boolean;
   public
    SubMenu           : array of TSubMenu;
    ActiveMenu        : Integer;
    MousePos          : TPoint;
    // Vars for new game screen
    SelAirCraft       : array[0..1] of ShortInt;
    SelScenario       : Integer;
    DC                : HDC;
    property Visible : Boolean read FVisible write FVisible;
    function AddSubMenu(pName, pBackGroundTexture : String) : TSubMenu;
    function AddMenuItem(pSubMenu : TSubMenu; pActiveTexture, pAction : String; pPosition, pSize : TSPoint) : TMenuItem;
    procedure SetActiveSubMenu(pName : String);
    procedure Render;
    procedure ProcessInput;
    procedure ProcessMenuAction(pMenuEntry : TMenuItem);
    constructor Create(pOwner : TComponent; pFontTexture : String; pDC : HDC); reintroduce;
    destructor Destroy; override;
  end;

 // TCloudLayer ================================================================
 TCloudLayer = class
   CloudTexture : array of glUInt;
   Cloud        : array of TVector3f;
   CloudsDrawn  : Integer;
   procedure Generate(pNumberOfClouds : Integer);
   procedure AddCloudTexture(pTextureFile : String);
   procedure Render(pViewerPos : TVector3f);
   destructor Destroy; override;
  end;

 // TScenarioSetting ===========================================================
 TScenarioSetting = record
   SkySphereTexID  : glUInt;
   SkySphereListID : glUInt;
   FogColor        : array[0..2] of Single;
   FogStart        : Single;
   FogEnd          : Single;
   SunPos          : TVector3f;
   SizeScale       : Single;
   HeightScale     : Single;
   WaterPlane      : Boolean;
   TerrainOffset   : Integer;
   TreeModel       : TAll3DSMesh;
  end;

 // TBonusCrate ================================================================
 TBonusCrate = class
   Bonus      : Integer;
   NewtonBody : PNewtonBody;
   Matrix     : TMatrix4f;
   Remove     : Boolean;
   procedure Render;
   procedure Apply(pAirCraft : TAirCraft);
   destructor Destroy; override;
  end;

 // TGameClass =================================================================
 TGameClass = class
   NewtonWorld     : PNewtonWorld;
   AccTimeSlice    : Single;
   TimeLastFrame   : Cardinal;
   OwnerForm       : TForm;
   SceneRotation   : TVector3f;
   SceneZoom       : Single;
   Scenario        : TScenarioSetting;
   TextureFont     : TTexFont;
   SplitScreen     : Boolean;
   ShowFPS         : Boolean;
   Pause           : Boolean;
   TimeFactor      : Single;
   DC              : HDC;
   Fade            : Single;
   Winner          : Byte;
   CrateModel      : TAll3DSMesh;
   BonusCrateTimer : Single;
   Quit            : Boolean;
   StartTick       : Cardinal;
   Frames          : Integer;
   // Network
   IsServer        : Boolean;
   IsClient        : Boolean;
   WaitForClient   : Boolean;
   WaitForServer   : Boolean;
   InGame          : Boolean;
   ReflectionTexID : array[0..2] of glUInt;
   ReflectionTimer : ShortInt;
   NetworkTimer    : Single;
   // Game objects
   AirCraft        : TAirCraftArray;
   AirCraftPreview : array of glUInt;
   AirCraftInfo    : array of record
                      Name      : String;
                      Speed     : Byte;
                      FirePower : Byte;
                     end;
   ScenarioPreview : array of glUInt;
   ScenarioName    : array of String;
   BonusCrate      : array of TBonusCrate;
   Tree            : array of record
                      NewtonBody : PNewtonBody;
                      Position   : TVector3f;
                      Radius     : Single;
                      Rotation   : Single;
                      Scale      : Single;
                     end;
   // Textures
   TextureName     : THashedStringList;
   TextureID       : array of glUInt;
   constructor Create(pOwnerForm : TForm; pNewtonWorld : PNewtonWorld; pDC : HDC);
   destructor Destroy; override;
   procedure LoadPreviewImages;
   procedure DrawQuad(pLeft, pRight, pTop, pBottom, pDepth : Single); overload;
   procedure DrawQuad(pLeft, pRight, pTop, pBottom, pDepth, pSStart, pSEnd, pTStart, pTEnd : Single); overload;
   procedure UpdateLoadingScreen(pMessage : String);
   procedure RenderInfo;
   procedure RenderHUD(pPlane, pEnemy : Integer;pOrthoWidth, pOrthoHeight : Integer);
   procedure SetCamera(pFollow : Integer);
   procedure RenderReflection(pFollow, pReflectionTextureIndex : Integer);
   procedure RenderTrees(pFollow : Integer);
   procedure Render(pFollow, pEnemy, pReflectionTexture : Integer);
   procedure GameLoop(pDC : HDC);
   procedure LoadScenario(pDirectory : String);
   procedure StartNewGame(pScenario : String; pPlane1, pPlane2 : Byte; pDC : HDC);
   procedure EndGame;
   // Texture pool
   procedure AddTexturesInDir(pDirectory : String; pUseMipMap : Boolean = True);
   function BindTexture(pName : String; pTarget : glUInt = GL_TEXTURE_2D) : glUInt;
   // Network
   procedure StartServer;
   procedure EndServer;
   procedure StartClient;
   procedure EndClient;
   procedure CheckNetwork;
   // Bonus objects
   procedure SpawnBonusCrate;
  end;

var
 GameMenu    : TGameMenu;
 GameClass   : TGameClass;
 CloudLayer  : TCloudLayer;

 ShowDebug   : Boolean;

 FPS         : Single;
 QPCFreq     : Int64;

 Log         : TStringList;

 Settings    : record
               Resolution   : TPoint;
               Frequency    : Word;
               Fullscreen   : Boolean;
               Sound        : Boolean;
               Music        : Boolean;
               Reflections  : Boolean;
               DrawDistance : Integer;
               Controller1  : String;
               Controller2  : String;
               TCPPort      : Word;
               end;

const
 PhysicsSleep     = 12;
 zNear            = 1;
 zFar             = 4096;
 FOV              = 60;
 ReflectionSize   = 512; // must be n^2 and smaller than the clientheight
 NetworkInterval  = 5;

function SPoint(pX,pY : Single) : TSPoint;

implementation

function SPoint(pX,pY : Single) : TSPoint;
begin
Result.x := pX;
Result.Y := pY;
end;

// =============================================================================
//  CollisionCrateAirCraft
// =============================================================================
//  Get's called when an aircraft collides with a dropped bonus crate
// =============================================================================
function CollisionCrateAirCraft(const material : PNewtonMaterial; const contact : PNewtonContact) : Integer; cdecl;
var
 AirCraft : TAirCraft;
 Crate    : TBonusCrate;
 i        : Integer;
begin
Result     := 1;
Crate      := nil;
Aircraft   := nil;
if Length(GameClass.BonusCrate) > 0 then
 for i := 0 to High(GameClass.BonusCrate) do
  begin
  if GameClass.BonusCrate[i].NewtonBody = ContactInfo.Body0 then
   Crate := GameClass.BonusCrate[i];
  if GameClass.BonusCrate[i].NewtonBody = ContactInfo.Body1 then
   Crate := GameClass.BonusCrate[i];
  end;
if Length(GameClass.AirCraft) > 0 then
 for i := 0 to High(GameClass.AirCraft) do
  begin
  if GameClass.AirCraft[i].NewtonBody = ContactInfo.Body0 then
   AirCraft := GameClass.AirCraft[i];
  if GameClass.AirCraft[i].NewtonBody = ContactInfo.Body1 then
   AirCraft := GameClass.AirCraft[i];
  end;
if (not Assigned(AirCraft)) or (not Assigned(Crate)) then
 exit;
Crate.Apply(AirCraft);
Crate.Remove := True;
Result       := 0;
end;

// =============================================================================
//  CollisionCrateTerrain
// =============================================================================
//  Get's called when a dropped bonus crate hits the terrain
// =============================================================================
function CollisionCrateTerrain(const material : PNewtonMaterial; const contact : PNewtonContact) : Integer; cdecl;
var
 Crate : TBonusCrate;
 i     : Integer;
begin
Crate := nil;
if Length(GameClass.BonusCrate) > 0 then
 for i := 0 to High(GameClass.BonusCrate) do
  begin
  if GameClass.BonusCrate[i].NewtonBody = ContactInfo.Body0 then
   Crate := GameClass.BonusCrate[i];
  if GameClass.BonusCrate[i].NewtonBody = ContactInfo.Body1 then
   Crate := GameClass.BonusCrate[i];
  end;
if Assigned(Crate) then
 begin
 if not Crate.Remove then
  for i := 0 to 3+Random(3) do
   ExplosionContainer.Spawn(V3(Crate.Matrix[3,0]+Random-Random, Crate.Matrix[3,1]+Random-Random, Crate.Matrix[3,2]+Random-Random), 6+Random(8));
 Crate.Remove := True;
 end;
Result := 1;
end;


// *****************************************************************************
//  TGameMenu
// *****************************************************************************

// =============================================================================
//  TGameMenu.Create
// =============================================================================
constructor TGameMenu.Create(pOwner : TComponent;pFontTexture : String; pDC : HDC);
begin
inherited Create(pOwner);
TextureFont := TTexFont.Create(pFontTexture, 'data\font1.fnt', 2);
DC := pDC;
SelAirCraft[0] := 0;
SelAirCraft[1] := 0;
SelScenario    := 0;
end;

// =============================================================================
//  TGameMenu.Destroy
// =============================================================================
destructor TGameMenu.Destroy;
var
 i : Integer;
begin
if Length(SubMenu) > 0 then
 for i := 0 to High(SubMenu) do
  SubMenu[i].Free;
TextureFont.Free;
inherited;
end;

// =============================================================================
//  TGameMenu.AddSubMenu
// =============================================================================
function TGameMenu.AddSubMenu(pName, pBackGroundTexture : String) : TSubMenu;
begin
SetLength(SubMenu, Length(SubMenu)+1);
SubMenu[High(SubMenu)] := TSubMenu.Create;
SubMenu[High(SubMenu)].Name              := pName;
SubMenu[High(SubMenu)].BackGroundTexture := pBackGroundTexture;
Result := SubMenu[High(SubMenu)];
end;

// =============================================================================
//  TGameMenu.AddMenuItem
// =============================================================================
function TGameMenu.AddMenuItem(pSubMenu : TSubMenu; pActiveTexture, pAction : String; pPosition, pSize : TSPoint) : TMenuItem;
begin
SetLength(pSubMenu.MenuItem, Length(pSubMenu.MenuItem)+1);
with pSubMenu.MenuItem[High(pSubMenu.MenuItem)] do
 begin
 ActiveTexture := pActiveTexture;
 Action        := pAction;
 Position      := pPosition;
 Size          := pSize;
 end;
end;

// =============================================================================
//  TGameMenu.SetActiveSubMenu
// =============================================================================
procedure TGameMenu.SetActiveSubMenu(pName : String);
var
 i : Integer;
begin
if Length(SubMenu) > 0 then
 for i := 0 to High(SubMenu) do
  if SubMenu[i].Name = pName then
   begin
   ActiveMenu := i;
   exit;
   end;
end;

// =============================================================================
//  TGameMenu.Render
// =============================================================================
procedure TGameMenu.Render;
const
 XPos : array[0..1] of Single = (192, 600);
var
 i,j  : Integer;
begin
if not Visible then
 exit;
glDisable(GL_CULL_FACE);
glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
glColor3f(1,1,1);
glDisable(GL_LIGHTING);
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
glOrtho(0, 800, 600, 0, -1,1);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
glDepthFunc(GL_ALWAYS);
GameClass.BindTexture(SubMenu[ActiveMenu].BackGroundTexture, GL_TEXTURE_2D);
GameClass.DrawQuad(0, 800, 600, 0, 0);
if SubMenu[ActiveMenu].Name = 'newgame' then
 begin
 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 glBindTexture(GL_TEXTURE_2D, GameClass.ScenarioPreview[SelScenario]);
 GameClass.DrawQuad(400-320, 400+320, 360+240, 360-240, 0.2);
 glBindTexture(GL_TEXTURE_2D, GameClass.AirCraftPreview[SelAircraft[0]]);
 GameClass.DrawQuad(198+160, 198-160, 256+120, 256-120, 0.3);
 glBindTexture(GL_TEXTURE_2D, GameClass.AirCraftPreview[SelAircraft[1]]);
 GameClass.DrawQuad(596-160, 596+160, 256+120, 256-120, 0.3);
 // Aircraft infos
 GameClass.BindTexture('aircraftinfobg');
 GameClass.DrawQuad(192-97.5, 192+97.5, 154+27.5, 154-27.5, 0.4);
 GameClass.DrawQuad(600-97.5, 600+97.5, 154+27.5, 154-27.5, 0.4);
 GameClass.BindTexture('aircraftinfostar');
 for i := 0 to 1 do
  begin
  if GameClass.AirCraftInfo[SelAircraft[i]].Speed > 0 then
   for j := 0 to GameClass.AirCraftInfo[SelAircraft[i]].Speed-1 do
    GameClass.DrawQuad(XPos[i]-97.5+96+j*19-16, XPos[i]-97.5+96+j*19+16, 154-27.5+14+16, 154-27.5+14-16, 0.4);
  if GameClass.AirCraftInfo[SelAircraft[i]].FirePower > 0 then
   for j := 0 to GameClass.AirCraftInfo[SelAircraft[i]].FirePower-1 do
    GameClass.DrawQuad(XPos[i]-97.5+96+j*19-16, XPos[i]-97.5+96+j*19+16, 154-27.5+39+16, 154-27.5+39-16, 0.4);
  end;
 glDisable(GL_BLEND);
 end;
if SubMenu[ActiveMenu].Name = 'gameover' then
 begin
 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 GameClass.BindTexture('winnerpl'+IntToStr(GameClass.Winner+1));
 GameClass.DrawQuad(400-154.5, 400+154.5, 300+120, 300-120, 0.3);
 glDisable(GL_BLEND);
 end;
if Length(SubMenu[ActiveMenu].MenuItem) > 0 then
 for i := 0 to High(SubMenu[ActiveMenu].MenuItem) do
  with SubMenu[ActiveMenu].MenuItem[i] do
   if (MousePos.X > Position.X-Size.X / 2) and (MousePos.X < Position.X+Size.X / 2) and
      (MousePos.Y > Position.Y-Size.Y / 2) and (MousePos.Y < Position.Y+Size.Y / 2) then
       begin
       glEnable(GL_BLEND);
       GameClass.BindTexture(ActiveTexture, GL_TEXTURE_2D);
       GameClass.DrawQuad(Position.X-Size.X/2, Position.X+Size.X/2, Position.Y+Size.Y/2, Position.Y-Size.Y/2, 0.5);
       glDisable(GL_BLEND);
       break;
       end;
//Cursor
glEnable(GL_BLEND);
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
GameClass.BindTexture('cursor');
GameClass.DrawQuad(MousePos.X, MousePos.X+48, MousePos.Y+48, MousePos.Y, 0.9);
glDisable(GL_BLEND);
glDepthFunc(GL_LESS);
glEnable(GL_CULL_FACE);
SwapBuffers(DC);
end;

// =============================================================================
//  TGameMenu.ProcessMenuAction
// =============================================================================
procedure TGameMenu.ProcessMenuAction(pMenuEntry : TMenuItem);
begin
SoundSystem.PlaySample('menuclick', False);
if pMenuEntry.Action = 'quit' then
 GameClass.OwnerForm.Close;
if pMenuEntry.Action = 'prevaircraft_p1' then
 begin
 dec(SelAirCraft[0]);
 if SelAirCraft[0] < 0 then
  SelAirCraft[0] := High(GameClass.AirCraftPreview);
 end;
if pMenuEntry.Action = 'nextaircraft_p1' then
 begin
 inc(SelAirCraft[0]);
 if SelAirCraft[0] > High(GameClass.AirCraftPreview) then
  SelAirCraft[0] := 0
 end;
if pMenuEntry.Action = 'prevaircraft_p2' then
 begin
 dec(SelAirCraft[1]);
 if SelAirCraft[1] < 0 then
  SelAirCraft[1] := High(GameClass.AirCraftPreview);
 end;
if pMenuEntry.Action = 'nextaircraft_p2' then
 begin
 inc(SelAirCraft[1]);
 if SelAirCraft[1] > High(GameClass.AirCraftPreview) then
  SelAirCraft[1] := 0
 end;
if pMenuEntry.Action = 'prevscenario' then
 begin
 dec(SelScenario);
 if SelScenario < 0 then
  SelScenario := High(GameClass.ScenarioPreview);
 end;
if pMenuEntry.Action = 'nextscenario' then
 begin
 inc(SelScenario);
 if SelScenario > High(GameClass.ScenarioPreview) then
  SelScenario := 0
 end;
// Start head-to-head game (splitscreen)
if pMenuEntry.Action = 'startgame' then
 begin
 SoundSystem.PlaySample('startgame', False);
 if not GameClass.IsServer then
  GameClass.StartNewGame(GameClass.ScenarioName[SelScenario], SelAircraft[0], SelAircraft[1], DC)
 else
  GameClass.StartServer;
 end;
// Start a TCP-Server
if pMenuEntry.Action = 'menu_startserver' then
 begin
 GameClass.IsServer := True;
 SetActiveSubMenu('newgame');
 exit;
 end;
// Try to connect to TCP-Server as a client
if pMenuEntry.Action = 'menu_startclient' then
 begin
 GameClass.StartClient;
 exit;
 end;
if pMenuEntry.Action = 'menu_newgame' then
 begin
 GameClass.IsServer := False;
 GameClass.IsClient := False;
 end;
if Pos('menu_', pMenuEntry.Action) = 1 then
 SetActiveSubMenu(Copy(pMenuEntry.Action, Length('menu_')+1, Length(pMenuEntry.Action)));
end;

// =============================================================================
//  TGameMenu.ProcessInput
// =============================================================================
procedure TGameMenu.ProcessInput;
var
 i : Integer;
begin
if not Visible then
 exit;
if Length(SubMenu[ActiveMenu].MenuItem) > 0 then
 for i := 0 to High(SubMenu[ActiveMenu].MenuItem) do
  with SubMenu[ActiveMenu].MenuItem[i] do
   if (MousePos.X > Position.X-Size.X / 2) and (MousePos.X < Position.X+Size.X / 2) and
      (MousePos.Y > Position.Y-Size.Y / 2) and (MousePos.Y < Position.Y+Size.Y / 2) then
       begin
       ProcessMenuAction(SubMenu[ActiveMenu].MenuItem[i]);
       break;
       end;
end;

// *****************************************************************************
//  TGameClass
// *****************************************************************************

// =============================================================================
//  TGameClass.Create
// =============================================================================
constructor TGameClass.Create(pOwnerForm : TForm; pNewtonWorld : PNewtonWorld; pDC : HDC);
begin
NewtonWorld        := pNewtonWorld;
OwnerForm          := pOwnerForm;
TimeLastFrame      := TimeGetTime;
SceneZoom          := -20;
SceneRotation      := V3(45, 45, 0);
gAirCraftArray     := @AirCraft;
TextureFont        := TTexFont.Create('data\font2.tga', 'data\font2.fnt', 2.2);
TextureName        := THashedStringList.Create;
ExplosionContainer := TExplosionContainer.Create;
DC                 := pDC;
LensFlare          := TLensFlare.Create;
ShowFPS            := False;
QueryPerformanceFrequency(QPCFreq);
glGenTextures(1, @ReflectionTexID[0]);
glBindTexture(GL_TEXTURE_2D, ReflectionTexID[0]);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, ReflectionSize, ReflectionSize, 0);
glGenTextures(1, @ReflectionTexID[1]);
glBindTexture(GL_TEXTURE_2D, ReflectionTexID[1]);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, ReflectionSize, ReflectionSize, 0);
glGenTextures(1, @ReflectionTexID[2]);
glBindTexture(GL_TEXTURE_2D, ReflectionTexID[2]);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, ReflectionSize, ReflectionSize, 0);
end;

// =============================================================================
//  TGameClass.Destroy
// =============================================================================
destructor TGameClass.Destroy;
var
 i : Integer;
begin
LensFlare.Free;
if Length(AirCraft) > 0 then
 for i := 0 to High(AirCraft) do
  if Assigned(AirCraft[i]) then
   AirCraft[i].Free;
if Length(BonusCrate) > 0 then
 for i := 0 to High(BonusCrate) do
  if Assigned(BonusCrate[i]) then
   BonusCrate[i].Free;
if Assigned(TextureName) then
 TextureName.Free;
if Assigned(TextureFont) then
 TextureFont.Free;
if Assigned(CloudLayer) then
 CloudLayer.Free;
if Assigned(Terrain) then
 Terrain.Free;
if Assigned(CrateModel) then
 CrateModel.Free;
end;

// =============================================================================
//  TGameClass.LoadPreviewImages
// =============================================================================
procedure TGameClass.LoadPreviewImages;
var
 CurrDir : String;
 SR      : TSearchRec;
 Ini     : TIniFile;
begin
// Preview images for the planes
CurrDir := GetCurrentDir;
ChDir(ExtractFilePath(Application.ExeName)+'data\planes');
if FindFirst('*.tga', faAnyFile, SR) = 0 then
 repeat
 SetLength(AirCraftPreview, Length(AirCraftPreview)+1);
 LoadTexture(SR.Name, AirCraftPreview[High(AirCraftPreview)]);
 SetLength(AirCraftInfo, Length(AirCraftInfo)+1);
 AirCraftInfo[High(AirCraftInfo)].Name := Copy(ExtractFileName(SR.Name), 1, Length(ExtractFileName(SR.Name))-4);;
 Ini := TIniFile.Create(ExtractFilePath(Application.ExeName)+'\data\planes\'+AirCraftInfo[High(AirCraftInfo)].Name+'.ini');
 AirCraftInfo[High(AirCraftInfo)].Speed     := Ini.ReadInteger('performance', 'speed', 1);
 AirCraftInfo[High(AirCraftInfo)].FirePower := Ini.ReadInteger('performance', 'firepower', 1);
 Ini.Free;
 until FindNext(SR) <> 0;
FindClose(SR);
ChDir(CurrDir);
// Preview images for the scenarios
ChDir(ExtractFilePath(Application.ExeName)+'data\scenarios');
if FindFirst('*.tga', faAnyFile, SR) = 0 then
 repeat
 SetLength(ScenarioPreview, Length(ScenarioPreview)+1);
 LoadTexture(SR.Name, ScenarioPreview[High(ScenarioPreview)]);
 SetLength(ScenarioName, Length(ScenarioName)+1);
 ScenarioName[High(ScenarioName)] := Copy(ExtractFileName(SR.Name), 1, Length(ExtractFileName(SR.Name))-4);;
 until FindNext(SR) <> 0;
FindClose(SR);
ChDir(CurrDir);
end;

// =============================================================================
//  TGameClass.DrawQuad
// =============================================================================
procedure TGameClass.DrawQuad(pLeft, pRight, pTop, pBottom, pDepth : Single);
begin
glBegin(GL_QUADS);
 glTexCoord2f(0,0); glVertex3f(pLeft,   pTop,    pDepth);
 glTexCoord2f(0,1); glVertex3f(pLeft,   pBottom, pDepth);
 glTexCoord2f(1,1); glVertex3f(pRight,  pBottom, pDepth);
 glTexCoord2f(1,0); glVertex3f(pRight,  pTop,    pDepth);
glEnd;
end;

procedure TGameClass.DrawQuad(pLeft, pRight, pTop, pBottom, pDepth, pSStart, pSEnd, pTStart, pTEnd : Single);
begin
glBegin(GL_QUADS);
 glTexCoord2f(pSStart, pTStart); glVertex3f(pLeft,   pTop,    pDepth);
 glTexCoord2f(pSStart, pTEnd);   glVertex3f(pLeft,   pBottom, pDepth);
 glTexCoord2f(pSEnd,   pTEnd);   glVertex3f(pRight,  pBottom, pDepth);
 glTexCoord2f(pSEnd,   pTStart); glVertex3f(pRight,  pTop,    pDepth);
glEnd;
end;

// =============================================================================
//  TGameClass.UpdateLoadingScreen
// =============================================================================
procedure TGameClass.UpdateLoadingScreen(pMessage : String);
begin
glDisable(GL_CULL_FACE);
glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
glColor3f(1,1,1);
glDisable(GL_LIGHTING);
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
glOrtho(0, 800, 600, 0, -1,1);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
BindTexture('loading', GL_TEXTURE_2D);
glBegin(GL_QUADS);
 glTexCoord2f(1,1); glVertex3f(800, 0,   0);
 glTexCoord2f(1,0); glVertex3f(800, 600, 0);
 glTexCoord2f(0,0); glVertex3f(0,   600, 0);
 glTexCoord2f(0,1); glVertex3f(0,   0,   0);
glEnd;
TextureFont.Print2DCenter(pMessage, [400,360,0.1], 2, False);
glEnable(GL_CULL_FACE);
SwapBuffers(DC);
end;

// =============================================================================
//  TGameClass.AddTexturesInDir
// =============================================================================
procedure TGameClass.AddTexturesInDir(pDirectory : String; pUseMipMap : Boolean = True);
var
 SR   : TSearchRec;
 FExt : String;
begin
Textures.UseMipMap := pUseMipMap;
if FindFirst(pDirectory+'*.*', faAnyFile, SR) = 0 then
 begin
 repeat
 FExt := ExtractFileExt(SR.Name);
 if (FExt = '.jpg') or (FExt = '.bmp') or (FExt = '.tga') then
  begin
  SetLength(TextureID, Length(TextureID)+1);
  LoadTexture(pDirectory+SR.Name, TextureID[High(TextureID)]);
  TextureName.Add(Copy(ExtractFileName(SR.Name), 1, Length(ExtractFileName(SR.Name))-4));
  end;
 until FindNext(SR) <> 0;
 end
else
 begin
 raise Exception.Create('TGameClass.AddTexturesInDir->Could not parse '+pDirectory+' (empty?)!');
 FindClose(SR);
 Application.Terminate;
 end;
FindClose(SR);
end;

// =============================================================================
//  TGameClass.BindTexture
// =============================================================================
function TGameClass.BindTexture(pName : String; pTarget : glUInt = GL_TEXTURE_2D) : glUInt;
var
 Index : Integer;
begin
Index := TextureName.IndexOf(pName);
if Index < 0 then
 begin
 raise Exception.Create('TGameClass.BindTexture->Texture '+pName+' not in texture pool!');
 Application.Terminate;
 end;
glBindTexture(pTarget, TextureID[Index]);
glEnable(pTarget);
Result := TextureID[Index];
end;

// =============================================================================
//  TGameClass.RenderInfo
// =============================================================================
procedure TGameClass.RenderInfo;
begin
glDisable(GL_LIGHTING);
glDisable(GL_DEPTH_TEST);
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
glOrtho(0, 640, 480, 0, -1,1);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
glEnable(GL_TEXTURE_2D);
TextureFont.Print2D(IntToStr(Round(FPS))+' FPS', [5,16,0], [1,1,0], 0.9, 2, False);
glEnable(GL_DEPTH_TEST);
end;

// =============================================================================
//  TGameClass.RenderHUD
// =============================================================================
procedure TGameClass.RenderHUD(pPlane, pEnemy : Integer; pOrthoWidth, pOrthoHeight : Integer);
var
 Viewport : TGLVectori4;
 Aspect   : Single;
 TmpS     : Single;
 Bottom   : Integer;
 i        : Integer;
 Factor   : Single;
begin
glDisable(GL_LIGHTING);
glDisable(GL_DEPTH_TEST);
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
glGetIntegerv(GL_VIEWPORT, @Viewport[0]);
glOrtho(0, pOrthoWidth, pOrthoHeight, 0, -1,1);
Bottom := pOrthoHeight;
Aspect := (640/480)/(ViewPort[2]/ViewPort[3]);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
glEnable(GL_BLEND);
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
glDisable(GL_TEXTURE_2D);
// Draw some overlay if underwater
AirCraft[pPlane].UnderWater := False;
if Scenario.WaterPlane then
 if AirCraft[pPlane].GetPosition[1] < 0 then
  begin
  glDepthMask(False);
  glColor4f(0, 0.3, 0.9, 0.75);
  DrawQuad(0, pOrthoWidth, pOrthoHeight, 0, 0);
  glDepthMask(True);
  glColor3f(1,1,1);
  AirCraft[pPlane].UnderWater := True;
  end;
// HUD =========================================================================
// Base
BindTexture('hud', GL_TEXTURE_2D);
glEnable(GL_TEXTURE_2D);
DrawQuad(-2, 254, Bottom+2, Bottom-254/Aspect, 0);
glPushMatrix;
 glTranslatef(-1,1,0);
 // Throttle
 BindTexture('hud_throttle_temp', GL_TEXTURE_2D);
 with AirCraft[pPlane] do
  DrawQuad(0, 160/2, Bottom, Bottom-256/Aspect*Throttle*0.6, 0,  0,160/512,0,1*Throttle*0.6);
 // MG temperature
 with AirCraft[pPlane] do
  DrawQuad(160/2, 160/2+256-160/2, Bottom, Bottom-256/Aspect*(GunHeat/140)*0.6, 0,  160/512,1,0,1*(GunHeat/140)*0.6);
 // Life
 BindTexture('life', GL_TEXTURE_2D);
 if AirCraft[pPlane].Health > 0 then
  with AirCraft[pPlane] do
   DrawQuad(79-64*(Health/100), 79+64*(Health/100), Bottom-75/Aspect-64/Aspect*(Health/100), Bottom-75/Aspect+64/Aspect*(Health/100), 0);
glPopMatrix;
// Draw hit indicator
if AirCraft[pPlane].HitTimer > 0 then
 begin
 glDisable(GL_TEXTURE_2D);
 glEnable(GL_BLEND);
 glColor4f(1,0,0,0.55);
 DrawQuad(0, pOrthoWidth, pOrthoHeight, 0, 0.1);
 glColor3f(1,1,1);
 glDisable(GL_BLEND);
 end;
// RADAR =======================================================================
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
glOrtho(0, Viewport[2], Viewport[3], 0, -1,1);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
glEnable(GL_BLEND);
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
glColor4f(1,1,1,0.5);
glPushMatrix;
 glTranslatef(Viewport[2]-96, Viewport[3]-96, 0);
 glRotatef(45, 0,0,1);
 glPushMatrix;
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, Terrain.BaseTexture);
  glBegin(GL_QUADS);
   glTexCoord2f(0,1); glVertex3f(-64, -64, 0);
   glTexCoord2f(0,0); glVertex3f(-64,  64, 0);
   glTexCoord2f(1,0); glVertex3f( 64,  64, 0);
   glTexCoord2f(1,1); glVertex3f( 64, -64, 0);
  glEnd;
 glPopMatrix;
 glDisable(GL_TEXTURE_2D);
 // Render markers
 // Own aircraft
 glPushMatrix;
  TmpS := arctan2(AirCraft[pPlane].Matrix[0,0], AirCraft[pPlane].Matrix[0,2]) * 180/Pi;
  if TmpS < 0 then
   TmpS := TmpS + 360;
  glTranslatef(AirCraft[pPlane].GetPosition[0]*128/Terrain.Width, AirCraft[pPlane].GetPosition[2]*128/Terrain.Height, 0);
  glRotatef(-TmpS, 0,0,1);
  glBegin(GL_TRIANGLES);
   glColor3f(0,0.3,0); glVertex3f(-5, -5, 0);
   glColor3f(0,1,0);   glVertex3f( 0,  5, 0);
   glColor3f(0,0.3,0); glVertex3f( 5, -5, 0);
  glEnd;
 glPopMatrix;
 // Enemy aircraft
 glPushMatrix;
  TmpS := arctan2(AirCraft[pEnemy].Matrix[0,0], AirCraft[pEnemy].Matrix[0,2]) * 180/Pi;
  if TmpS < 0 then
   TmpS := TmpS + 360;
  glTranslatef(AirCraft[pEnemy].GetPosition[0]*128/Terrain.Width, AirCraft[pEnemy].GetPosition[2]*128/Terrain.Height, 0);
  glRotatef(-TmpS, 0,0,1);
  glBegin(GL_TRIANGLES);
   glColor3f(0.3,0,0); glVertex3f(-5, -5, 0);
   glColor3f(1,0,0);   glVertex3f( 0,  5, 0);
   glColor3f(0.3,0,0); glVertex3f( 5, -5, 0);
  glEnd;
 glPopMatrix;
 glColor4f(1,0,0,0.6);
 // Bonus crates
 glColor4f(1,1,0,0.6);
 if Length(BonusCrate) > 0 then
  for i := 0 to High(BonusCrate) do
   begin
   glPushMatrix;
    glTranslatef(BonusCrate[i].Matrix[3,0]*128/Terrain.Width, BonusCrate[i].Matrix[3,2]*128/Terrain.Height, 0);
    DrawQuad(-2, 2, -2, 2, 0);
   glPopMatrix;
   end;
 glColor3f(1,1,1);
 glEnable(GL_TEXTURE_2D);
glPopMatrix;
// Lifes
for i := 0 to MaxLifes-1 do
 begin
 if AirCraft[pPlane].Lifes > i then
  BindTexture('iconlife')
 else
  BindTexture('iconlifeempty');
 DrawQuad(5+i*36, 5+i*36+32, 5+28, 5, 1);
 end;
Factor := (Viewport[2]/Viewport[3])/(1+1/3);
// Bonus
if AirCraft[pPlane].LastBonusTimer > 0 then
 begin
 BindTexture('bonus'+IntToStr(AirCraft[pPlane].LastBonus));
 glEnable(GL_BLEND);
 glColor4f(1,1,1, AirCraft[pPlane].LastBonusTimer/10);
 DrawQuad(Viewport[2]/2-128/Factor*10/AirCraft[pPlane].LastBonusTimer, Viewport[2]/2+128/Factor*10/AirCraft[pPlane].LastBonusTimer, Viewport[3]/2+128/Factor*10/AirCraft[pPlane].LastBonusTimer, Viewport[3]/2-128/Factor*10/AirCraft[pPlane].LastBonusTimer, 0);
 glColor3f(1,1,1);
 glDisable(GL_BLEND);
 end;
// Some states
if AirCraft[pPlane].GunHeat > 120 then
 begin
 glEnable(GL_BLEND);
 glColor4f(1,1,1,0.6);
 BindTexture('text_mgunstoohot', GL_TEXTURE_2D);
 DrawQuad(Viewport[2]/2-128/Factor, Viewport[2]/2+128/Factor, 140/Factor+128/Factor, 140/Factor-128/Factor, 0);
 glColor3f(1,1,1);
 glDisable(GL_BLEND);
 end;
if AirCraft[pPlane].OutOfLevel then
 begin
 glEnable(GL_BLEND);
 glColor4f(1,1,1,0.6);
 BindTexture('text_returntoscene', GL_TEXTURE_2D);
 DrawQuad(Viewport[2]/2-128/Factor, Viewport[2]/2+128/Factor, ViewPort[3]/2+128/Factor, ViewPort[3]/2-128/Factor, 0);
 glColor3f(1,1,1);
 glDisable(GL_BLEND);
 end;
glDisable(GL_BLEND);
glDisable(GL_TEXTURE_2D);
glEnable(GL_DEPTH_TEST);
end;

// =============================================================================
//  TGameClass.SetCamera
// =============================================================================
procedure TGameClass.SetCamera(pFollow : Integer);
var
 TmpM : TMatrix4f;
begin
case AirCraft[pFollow].Camera of
 ctFreeRotate :
  begin
  glTranslatef(0, 0, SceneZoom);
  glRotatef(SceneRotation[0], 1,0,0);
  glRotatef(SceneRotation[1], 0,1,0);
  glRotatef(SceneRotation[2], 0,0,1);
  TmpM := AirCraft[pFollow].Matrix;
  glTranslatef(-TmpM[3,0], -TmpM[3,1], -TmpM[3,2]);
  end;
 ctCloseBehind, ctFarBehind, ctAboveCockpit :
  begin
  case AirCraft[pFollow].Camera of
   ctCloseBehind  : glTranslatef(0, -3, -10);
   ctFarBehind    : glTranslatef(0, -1, -15);
   ctAboveCockpit : glTranslatef(0, -1.2, -2);
  end;
  TmpM := AirCraft[pFollow].Matrix;
  Matrix_Inverse(TmpM);
  glRotatef(90, 0,1,0);
  glMultMatrixf(@TmpM[0,0]);
  end;
end;
end;

// =============================================================================
//  TGameClass.RenderReflection
// =============================================================================
procedure TGameClass.RenderReflection(pFollow : Integer; pReflectionTextureIndex : Integer);
const
 CP   : array[0..3] of Double = (0,1,0,0);
var
 TmpM     : TMatrix4f;
 ViewPort : TGLVectori4;
 i        : Integer;
begin
if (not Settings.Reflections) or (not Scenario.WaterPlane) then
 exit;

glGetIntegerv(GL_VIEWPORT, @ViewPort);
glViewport(0, 0, ReflectionSize, ReflectionSize);
glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
gluPerspective(FOV, ViewPort[2]/ViewPort[3], zNear, zFar);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;

glPushMatrix;
 SetCamera(pFollow);
 glTranslatef(0, 1, 0);
 glScalef(1,-1,1);
 glClipPlane(GL_CLIP_PLANE0, @CP);
glPopMatrix;

// Render the skydome
glPushMatrix;
 case AirCraft[pFollow].Camera of
  ctFreeRotate :
   begin
   glRotatef(SceneRotation[0], 1,0,0);
   glRotatef(SceneRotation[1], 0,1,0);
   end;
  ctCloseBehind, ctFarBehind, ctAboveCockpit :
   begin
   TmpM := AirCraft[pFollow].Matrix;
   Matrix_SetTransform(TmpM, V3(0,0,0));
   Matrix_Inverse(TmpM);
   glRotatef(90, 0,1,0);
   glMultMatrixf(@TmpM[0,0]);
   end;
 end;
 glScalef(1, -1, 1);
 glCallList(Scenario.SkySphereListID);
glPopMatrix;

glEnable(GL_CLIP_PLANE0);

SetCamera(pFollow);
glScalef(1, -1, 1);

glFrontFace(GL_CW);

Frustum.Calculate;

// Render the terrain
glDisable(GL_LIGHTING);
glEnable(GL_FOG);
glFogi(GL_FOG_MODE, GL_LINEAR);
glFogf(GL_FOG_START, Scenario.FogStart);
glFogf(GL_FOG_END, Scenario.FogEnd);
glFogfv(GL_FOG_COLOR, @Scenario.FogColor);
Terrain.Render(AirCraft[pFollow].GetPosition, Settings.DrawDistance);
glDisable(GL_FOG);

// Render aircrafts
if Length(AirCraft) > 0 then
 for i := 0 to High(AirCraft) do
  AirCraft[i].Render;

// Bonus crates
if Length(BonusCrate) > 0 then
 for i := 0 to High(BonusCrate) do
  BonusCrate[i].Render;

RenderTrees(pFollow);

// Render explosions
BindTexture('explosion');
ExplosionContainer.Render;

glDisable(GL_CLIP_PLANE0);

glBindTexture(GL_TEXTURE_2D, ReflectionTexID[pReflectionTextureIndex]);
glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, ReflectionSize, ReflectionSize);

glViewPort(ViewPort[0], ViewPort[1], ViewPort[2], ViewPort[3]);
glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
glFrontFace(GL_CCW);
glDisable(GL_CULL_FACE);
end;

// =============================================================================
//  TGameClass.RenderTrees
// =============================================================================
procedure TGameClass.RenderTrees(pFollow : Integer);
var
 i : Integer;
begin
// Trees
glEnable(GL_FOG);
glEnable(GL_ALPHA_TEST);
glAlphaFunc(GL_GREATER, 0.75);
glEnable(GL_BLEND);
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
if Length(Tree) > 0 then
 for i := 0 to High(Tree) do
  if Abs(VDistance(Tree[i].Position, AirCraft[pFollow].GetPosition)) < Settings.DrawDistance then
   if Frustum.IsSphereWithin(Tree[i].Position[0], Tree[i].Position[1], Tree[i].Position[2], Tree[i].Radius) then
    begin
    glPushMatrix;
     glTranslatef(Tree[i].Position[0], Tree[i].Position[1], Tree[i].Position[2]);
     glScalef(Tree[i].Scale, Tree[i].Scale, Tree[i].Scale);
     glRotatef(Tree[i].Rotation, 0,1,0);
     glFrontFace(GL_CW);
      glCallList(Scenario.TreeModel.DisplayList);
     glFrontFace(GL_CCW);
    glPopMatrix;
    end;
glDisable(GL_BLEND);
glDisable(GL_ALPHA_TEST);
glDisable(GL_FOG);
end;

// =============================================================================
//  TGameClass.Render
// =============================================================================
procedure TGameClass.Render(pFollow, pEnemy, pReflectionTexture : Integer);
const
 FogWater  : array[0..2] of Single = (0, 0, 0.8);
var
 TmpM      : TMatrix4f;
 TmpV      : TVector3f;
 i         : Integer;
 Viewport  : TGLVectori4;
 SaveMat   : TMatrix4f;
begin
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;

// Render the skydome
glPushMatrix;
 case AirCraft[pFollow].Camera of
  ctFreeRotate :
   begin
   glRotatef(SceneRotation[0], 1,0,0);
   glRotatef(SceneRotation[1], 0,1,0);
   end;
  ctCloseBehind, ctFarBehind, ctAboveCockpit :
   begin
   TmpM := AirCraft[pFollow].Matrix;
   Matrix_SetTransform(TmpM, V3(0,0,0));
   Matrix_Inverse(TmpM);
   glRotatef(90, 0,1,0);
   glMultMatrixf(@TmpM[0,0]);
   end;
 end;
 glCallList(Scenario.SkySphereListID);
 glGetFloatv(GL_MODELVIEW_MATRIX, @SaveMat[0,0]);
glPopMatrix;

SetCamera(pFollow);
Frustum.Calculate;

// Render the terrain
glEnable(GL_FOG);
if AirCraft[pFollow].GetPosition[1] > 0 then
 begin
 glFogi(GL_FOG_MODE, GL_LINEAR);
 glFogf(GL_FOG_START, Scenario.FogStart);
 glFogf(GL_FOG_END, Scenario.FogEnd);
 glFogfv(GL_FOG_COLOR, @Scenario.FogColor);
 end;
Terrain.Render(AirCraft[pFollow].GetPosition, Settings.DrawDistance);
glDisable(GL_FOG);
// Water plane
glEnable(GL_BLEND);
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
if Settings.Reflections then
 begin
 glMatrixMode(GL_TEXTURE);
 glLoadIdentity;
 glTranslatef(0.5, 0.5, 0);
 glScalef(0.5, 0.5, 0);
 glGetIntegerv(GL_VIEWPORT, @ViewPort[0]);
 gluPerspective(FOV, ViewPort[2]/ViewPort[3], zNear, zFar);
 SetCamera(pFollow);
 glScalef(1, -1, 1);
 glMatrixMode(GL_MODELVIEW);
 glEnable(GL_TEXTURE_2D);
 glBindTexture(GL_TEXTURE_2D, ReflectionTexID[pReflectionTexture]);
 glColor4f(0.75, 0.75, 1, 0.8);
 glDepthMask(False);
 glBegin(GL_QUADS);
  glTexCoord3f(-Terrain.Width, 0, -Terrain.Height); glVertex3f(-Terrain.Width, 0, -Terrain.Height);
  glTexCoord3f(-Terrain.Width, 0,  Terrain.Height); glVertex3f(-Terrain.Width, 0,  Terrain.Height);
  glTexCoord3f( Terrain.Width, 0,  Terrain.Height); glVertex3f( Terrain.Width, 0,  Terrain.Height);
  glTexCoord3f( Terrain.Width, 0, -Terrain.Height); glVertex3f( Terrain.Width, 0, -Terrain.Height);
 glEnd;
 glDepthMask(True);
 glMatrixMode(GL_TEXTURE);
 glLoadIdentity;
 glMatrixMode(GL_MODELVIEW);
 glColor4f(1, 1, 1, 0.1);
 end
else
 glColor4f(1, 1, 1, 0.5);
BindTexture('water');
glBegin(GL_QUADS);
 glTexCoord2f(0,  0);  glVertex3f(-Terrain.Width, 0, -Terrain.Height);
 glTexCoord2f(0,  32); glVertex3f(-Terrain.Width, 0,  Terrain.Height);
 glTexCoord2f(32, 32); glVertex3f( Terrain.Width, 0,  Terrain.Height);
 glTexCoord2f(32, 0);  glVertex3f( Terrain.Width, 0, -Terrain.Height);
glEnd;
glColor3f(1,1,1);
glDisable(GL_BLEND);

RenderTrees(pFollow);

// Render aircrafts
if Length(AirCraft) > 0 then
 for i := 0 to High(AirCraft) do
  AirCraft[i].Render;

// Bonus crates
if Length(BonusCrate) > 0 then
 for i := 0 to High(BonusCrate) do
  BonusCrate[i].Render;

// Render cloud layer
CloudLayer.Render(AirCraft[pFollow].GetPosition);

// Render explosions
BindTexture('explosion');
ExplosionContainer.Render;

// Lens flare
glPushMatrix;
 glLoadMatrixf(@SaveMat[0,0]);
 glGetIntegerv(GL_VIEWPORT, @Viewport[0]);
 LensFlare.Render(Scenario.SunPos, AirCraft[pFollow].GetPosition, (Viewport[2]/Viewport[3])/(1+1/3), TimeFactor);
glPopMatrix;

// Render debug output
if ShowDebug then
 begin
 glColor3f(1,1,1);
 glDisable(GL_LIGHTING);
 glDisable(GL_TEXTURE_2D);
 glBegin(GL_LINES);
  NewtonWorldForEachBodyDo(NewtonWorld, @Debug_ShowBodyCollision);
 glEnd;
 g_Debug.RenderDebugLines;
 end;

// Targetindicator =============================================================
Frustum.Calculate;
if Frustum.IsPointWithin(AirCraft[pEnemy].GetPosition[0], AirCraft[pEnemy].GetPosition[1], AirCraft[pEnemy].GetPosition[2]) then
 if Abs(VDistance(AirCraft[pEnemy].GetPosition, AirCraft[pFollow].GetPosition)) > 150 then
  begin
  glGetFloatv(GL_MODELVIEW_MATRIX, @TmpM[0,0]);
  TmpV := Matrix_TansformVector(TmpM, AirCraft[pEnemy].GetPosition);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  BindTexture('targetindicator', GL_TEXTURE_2D);
  glPushMatrix;
   glLoadIdentity;
   GameClass.DrawQuad(TmpV[0]-10, TmpV[0]+10, TmpV[1]-10, TmpV[1]+10, TmpV[2]);
  glPopMatrix;
  glDisable(GL_BLEND);
  end;
end;

// =============================================================================
//  TGameClass.GameLoop
// =============================================================================
procedure TGameClass.GameLoop(pDC : HDC);
var
 QPCStart : Int64;
 QPCEnd   : Int64;
 i,j      : Integer;
 TmpCrate : TBonusCrate;
begin
repeat

// Network
CheckNetwork;

if IsServer then
 Pause := WaitForClient;
//if IsClient then
// Pause := WaitForServer;

// Menu
if Assigned(GameMenu) then
 if GameMenu.Visible then
  begin
  glViewPort(0, 0, OwnerForm.ClientWidth, OwnerForm.ClientHeight);
  Application.ProcessMessages;
  GameMenu.Render;
  continue;
  end;

// Timing (for update cycles etc.)
QueryPerformanceCounter(QPCStart);

// Accumulate time for physics since last frame
if not Pause then
 AccTimeSlice  := AccTimeSlice + (TimeGetTime-TimeLastFrame);
TimeLastFrame := TimeGetTime;

// Splitscreen
if SplitScreen then
 begin
 glViewPort(0, OwnerForm.ClientHeight-OwnerForm.ClientHeight div 2, OwnerForm.ClientWidth, OwnerForm.ClientHeight div 2);
 RenderReflection(0, 1);
 glViewPort(0, 0, OwnerForm.ClientWidth, OwnerForm.ClientHeight div 2);
 RenderReflection(1, 2);
 // Top (Player 1) =============================================================
 glEnable(GL_SCISSOR_TEST);
 glViewPort(0, OwnerForm.ClientHeight-OwnerForm.ClientHeight div 2, OwnerForm.ClientWidth, OwnerForm.ClientHeight div 2);
 glScissor(0, OwnerForm.ClientHeight-OwnerForm.ClientHeight div 2, OwnerForm.ClientWidth, OwnerForm.ClientHeight div 2);
 glClear(GL_DEPTH_BUFFER_BIT);
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity;
 gluPerspective(FOV, OwnerForm.ClientWidth/(OwnerForm.ClientHeight/2), zNear, zFar);
 Render(0,1,1);
 RenderHUD(0,1,1024,1024);
 // Bottom (Player 2) ==========================================================
 glEnable(GL_SCISSOR_TEST);
 glViewPort(0, 0, OwnerForm.ClientWidth, OwnerForm.ClientHeight div 2);
 glScissor(0, 0, OwnerForm.ClientWidth, OwnerForm.ClientHeight div 2);
 glClear(GL_DEPTH_BUFFER_BIT);
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity;
 gluPerspective(FOV, OwnerForm.ClientWidth/(OwnerForm.ClientHeight/2), zNear, zFar);
 Render(1,0,2);
 RenderHUD(1,0,1024,1024);
 glDisable(GL_SCISSOR_TEST);
 // Render a splitter ==========================================================
 glDisable(GL_SCISSOR_TEST);
 glViewPort(0, 0, OwnerForm.ClientWidth, OwnerForm.ClientHeight);
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity;
 glOrtho(0, 640, 0, 480, -1,1);
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity;
 BindTexture('splitter', GL_TEXTURE_2D);
 glEnable(GL_BLEND);
 DrawQuad(0,640,232,246,0);
 glDisable(GL_BLEND);
 end
else
 begin
 glClear(GL_DEPTH_BUFFER_BIT);
 RenderReflection(0,0);
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity;
 gluPerspective(FOV, OwnerForm.ClientWidth/OwnerForm.ClientHeight, zNear, zFar);
 glViewPort(0, 0, OwnerForm.ClientWidth, OwnerForm.ClientHeight);
 Render(0,1,0);
 RenderHUD(0,1,640,480);
 end;

if ShowFPS then
 RenderInfo;

// Fade out
if (Fade > 0) and (Fade < 1.5) then
 begin
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity;
 glOrtho(0, 1, 1, 0, -1, 1);
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity;
 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 glColor4f(0,0,0, Fade);
 glDisable(GL_TEXTURE_2D);
 glDepthFunc(GL_ALWAYS);
 GameClass.DrawQuad(0,1,0,1,0);
 glDepthFunc(GL_LESS);
 end;

SwapBuffers(pDC);

// Update physics
if not Pause then
 while AccTimeSlice > PhysicsSleep do
  begin
  NewtonUpdate(NewtonWorld, (PhysicsSleep / 1000));
  NewtonUpdate(NewtonWorld, (PhysicsSleep / 1000));
  AccTimeSlice := AccTimeSlice - PhysicsSleep;
  end;

// Timing (for update cycles etc.)
QueryPerformanceCounter(QPCEnd);
TimeFactor := ((QPCEnd-QPCStart)/QPCFreq)*100;

// Bonus drops
if (not Pause) and (not IsClient) then
 begin
 BonusCrateTimer := BonusCrateTimer - 0.1 * TimeFactor;
 if BonusCrateTimer < 0 then
  begin
  if (Random(3) = 0) and (Length(BonusCrate) < 5) then
   SpawnBonusCrate;
  BonusCrateTimer := 25;
  end;
 end;

// Fading out
if Fade < 1.5 then
 Fade := Fade + TimeFactor * 0.01;

// Game objects
if not Pause then
 ExplosionContainer.Update(TimeFactor);
if Length(BonusCrate) = 1 then
 if BonusCrate[0].Remove then
  begin
  if IsServer then
   NetworkClass.ServerSendPackage(nmServer_RemoveBonusCrate, '0', '', '');
  BonusCrate[0].Free;
  SetLength(BonusCrate, 0);
  end;
if Length(BonusCrate) > 1 then
 for i := 0 to High(BonusCrate) do
  if i < Length(BonusCrate) then
   if BonusCrate[i].Remove then
    begin
    TmpCrate                     := BonusCrate[High(BonusCrate)];
    BonusCrate[High(BonusCrate)] := BonusCrate[i];
    BonusCrate[i]                := TmpCrate;
    BonusCrate[High(BonusCrate)].Free;
    SetLength(BonusCrate, Length(BonusCrate)-1);
    if IsServer then
     NetworkClass.ServerSendPackage(nmServer_RemoveBonusCrate, IntToStr(i), '', '');
    end;
if Length(AirCraft) > 0 then
 for i := 0 to High(AirCraft) do
  begin
  AirCraft[i].Update(TimeFactor);
  if AirCraft[i].Lifes < 1 then
   begin
   if Fade = 1.5 then
    Fade := 0;
   if Fade > 1.1 then
    begin
    for j := 0 to High(AirCraft) do
     if AirCraft[j].Lifes > 0 then
      Winner := j;
    GameMenu.SetActiveSubMenu('gameover');
    GameMenu.Visible := True;
    end;
   end;
  end;

inc(Frames);
if GetTickCount - StartTick >= 500 then
 begin
 FPS       := Frames/(GetTickCount-StartTick)*1000;
 Frames    := 0;
 StartTick := GetTickCount
 end;

Application.ProcessMessages;

if Length(GameClass.AirCraft) > 0 then
 for i := 0 to High(GameClass.AirCraft) do
  GameClass.AirCraft[i].ProcessInput(TimeFactor);

until Quit;
end;

// =============================================================================
//  TGameClass.LoadScenario
// =============================================================================
procedure TGameClass.LoadScenario(pDirectory : String);
var
 Q        : Pointer;
 INI      : TIniFile;
 TmpPos   : TVector3f;
 TmpM     : TMatrix4f;
 i,j,f    : Integer;
 PosOK    : Boolean;
 Collider : PNewtonCollision;
 TmpFace  : array[0..2] of TVector3f;
begin
Log.Add(TimeToStr(Now)+':  Loading scenario "'+pDirectory+'"');
ChDir(ExtractFilePath(Application.ExeName)+pDirectory);
// Settings
if not FileExists(GetCurrentDir+'\settings.ini') then
 begin
 Log.Add(TimeToStr(Now)+':  TGameClass.LoadScenario->Scenario settings file not found!');
 raise Exception.Create('TGameClass.LoadScenario->Scenario settings file not found!');
 Application.Terminate;
 end;
INI := TIniFile.Create(GetCurrentDir+'\settings.ini');
Scenario.FogColor[0]   := INI.ReadInteger('scenario', 'fogred', 127)/255;
Scenario.FogColor[1]   := INI.ReadInteger('scenario', 'foggreen', 127)/255;
Scenario.FogColor[2]   := INI.ReadInteger('scenario', 'fogblue', 127)/255;
Scenario.FogStart      := INI.ReadInteger('scenario', 'fogstart', 256);
Scenario.FogEnd        := INI.ReadInteger('scenario', 'fogend', 768);
Scenario.SunPos[0]     := INI.ReadInteger('scenario', 'sunx', 512);
Scenario.SunPos[1]     := INI.ReadInteger('scenario', 'suny', 512);
Scenario.SizeScale     := INI.ReadInteger('scenario', 'sizescale', 100);
Scenario.HeightScale   := INI.ReadInteger('scenario', 'heightscale', 100);
Scenario.WaterPlane    := INI.ReadBool('scenario', 'waterplane', True);
Scenario.TerrainOffset := INI.ReadInteger('scenario', 'terrainoffset', -50);
Scenario.SizeScale   := Scenario.SizeScale/100;
Scenario.HeightScale := Scenario.HeightScale/100;
INI.Free;
UpdateLoadingScreen('Generating sky sphere...');
// Build sky sphere
LoadTexture('skysphere.jpg', Scenario.SkySphereTexID);
Scenario.SunPos[0] := DegToRad(Scenario.SunPos[0] / LastWidth * 360);
Scenario.SunPos[1] := DegToRad(Scenario.SunPos[1] / LastHeight * 180);
TmpPos := V3(512 * Cos(Scenario.SunPos[0]) * Sin(Scenario.SunPos[1]), -(512 * Sin(Scenario.SunPos[0]) * Sin(Scenario.SunPos[1])), 512 * Cos(Scenario.SunPos[1]));
Scenario.SunPos := V3(TmpPos[1], TmpPos[0], TmpPos[2]);
Scenario.SkySphereListID := glGenLists(1);
glNewList(Scenario.SkySphereListID, GL_COMPILE);
 Q := gluNewQuadric;
 gluQuadricTexture(Q, True);
 glColor3f(1,1,1);
 glDisable(GL_LIGHTING);
 glEnable(GL_TEXTURE_2D);
 glBindTexture(GL_TEXTURE_2D, Scenario.SkySphereTexID);
 glDepthMask(False);
 glRotatef(-90, 1,0,0);
 gluSphere(Q, 512, 16, 16);
 glDepthMask(True);
 gluDeleteQuadric(Q);
glEndList;
UpdateLoadingScreen('Generating cloud layer');
// Generate cloud layer
if Assigned(CloudLayer) then
 CloudLayer.Free;
CloudLayer := TCloudLayer.Create;
CloudLayer.AddCloudTexture(ExtractFilePath(Application.ExeName)+'data\textures\cloud.tga');
CloudLayer.Generate(500);
// Terrain
UpdateLoadingScreen('Loading terrain');
if Assigned(Terrain) then
 Terrain.Free;
Terrain := TTerrain.Create;
Terrain.LoadFromFile('terrain.bmp', 'terrainbase.jpg', 'terraindetail.tga', NewtonWorld, Scenario.SizeScale , Scenario.HeightScale, Scenario.TerrainOffset);
// Generate vegetation
UpdateLoadingScreen('Creating vegetation');
if Assigned(Scenario.TreeModel) then
 Scenario.TreeModel.Free;
Scenario.TreeModel := TAll3DSMesh.Create(nil);
Scenario.TreeModel.TexturePath := ExtractFilePath(Application.ExeName)+'data\textures\3ds\';
Scenario.TreeModel.LoadFromFile('tree.3ds');
Scenario.TreeModel.BuildDisplayList;
if Length(Tree) > 0 then
 for i := 0 to High(Tree) do
  if Assigned(Tree[i].NewtonBody) then
   NewtonDestroyBody(NewtonWorld, Tree[i].NewtonBody);
Collider := nil;
// Build tree collision for 3DS model of current scenario tree
SetLength(Tree, 250);
for i := 0 to High(Tree) do
 with Tree[i] do
  begin
  Scale       := 1+Random*2;
  Radius      := Scenario.TreeModel.Maximum.x-Scenario.TreeModel.Minimum.x;
  if Scenario.TreeModel.Maximum.y-Scenario.TreeModel.Minimum.y > Radius then
   Radius := Scenario.TreeModel.Maximum.y-Scenario.TreeModel.Minimum.y;
  if Scenario.TreeModel.Maximum.z-Scenario.TreeModel.Minimum.z > Radius then
   Radius := Scenario.TreeModel.Maximum.z-Scenario.TreeModel.Minimum.z;
  Radius   := Radius * Scale;
  Rotation := Random(360);
  repeat
  PosOK       := True;
  Position    := V3((Random-Random) * Terrain.Width/2 * 0.95, 600, (Random-Random) * Terrain.Height/2 * 0.95);
  Position[1] := GetHeight(NewtonWorld, Position);
  // Position below waterplane
  if (Position[1] < 0) or (Position[1] >= 1000) then
   PosOK := False;
  // Another tree on same position/within radius
  for j := 0 to High(Tree) do
   if i <> j then
    if Abs(VDistance(Tree[j].Position, Position)) < Radius then
     PosOK := False;
  until PosOK;
  Collider := NewtonCreateTreeCollision(NewtonWorld, nil);
  NewtonTreeCollisionBeginBuild(Collider);
  with Scenario.TreeModel.Mesh[0] do
   begin
   f := 0;
   repeat
   Move(FVertex[Face[f  ]], TmpFace[0][0], 3*SizeOf(Single));
   Move(FVertex[Face[f+1]], TmpFace[1][0], 3*SizeOf(Single));
   Move(FVertex[Face[f+2]], TmpFace[2][0], 3*SizeOf(Single));
   TmpFace[0] := VScale(TmpFace[0], Scale);
   TmpFace[1] := VScale(TmpFace[1], Scale);
   TmpFace[2] := VScale(TmpFace[2], Scale);
   NewtonTreeCollisionAddFace(Collider, 3, @TmpFace[0], SizeOf(TVector3f), 0);
   inc(f,3);
   until f > NumFaces-1;
   end;
  NewtonTreeCollisionEndBuild(Collider, 1);

  NewtonBody := NewtonCreateBody(NewtonWorld, Collider);
  Matrix_SetIdentity(TmpM);
  Matrix_SetRotation(TmpM, V3(0, -Rotation, 0));
  Matrix_SetTransform(TmpM, Position);
  NewtonBodySetMatrix(NewtonBody, @TmpM[0,0]);
  NewtonBodySetMaterialGroupID(NewtonBody, MatTerrain);
  end;
NewtonReleaseCollision(NewtonWorld, Collider);
// Update time for physics
TimeLastFrame := TimeGetTime;
ChDir(ExtractFilePath(Application.ExeName));
Log.Add(TimeToStr(Now)+':  Scenario loaded');
end;

// =============================================================================
//  TGameClass.StartNewGame
// =============================================================================
procedure TGameClass.StartNewGame(pScenario : String; pPlane1, pPlane2 : Byte; pDC : HDC);
var
 i : Integer;
 M : TMatrix4f;
begin
Log.Add(TimeToStr(Now)+': Starting new game');
GameClass.UpdateLoadingScreen('Cleaning up');
ExplosionContainer.Clear;
if Length(AirCraft) > 0 then
 for i := 0 to High(AirCraft) do
  AirCraft[i].Free;
if Length(BonusCrate) > 0 then
 for i := 0 to High(BonusCrate) do
  BonusCrate[i].Free;
SetLength(BonusCrate, 0);
// Save random seed and send to client later, so that the client also generates the same random tree positions
NetworkClass.ServerRandSeed := RandSeed;
// Generate newton materials
MatTerrain    := NewtonMaterialCreateGroupID(NewtonWorld);
MatAirCraft   := NewtonMaterialCreateGroupID(NewtonWorld);
MatBonusCrate := NewtonMaterialCreateGroupID(NewtonWorld);
// Set callbacks for collision between different material groups
NewtonMaterialSetCollisionCallback(NewtonWorld, MatTerrain,    MatAirCraft, @MatTerrain,    @CollisionBegin, @CollisionTerrainAirCraft,  nil);
NewtonMaterialSetCollisionCallback(NewtonWorld, MatAirCraft,   MatAirCraft, @MatAirCraft,   @CollisionBegin, @CollisionAirCraftAirCraft, nil);
NewtonMaterialSetCollisionCallback(NewtonWorld, MatBonusCrate, MatAirCraft, @MatBonusCrate, @CollisionBegin, @CollisionCrateAirCraft,    nil);
NewtonMaterialSetCollisionCallback(NewtonWorld, MatBonusCrate, MatTerrain , @MatBonusCrate, @CollisionBegin, @CollisionCrateTerrain,     nil);
// Load aircrafts
GameClass.UpdateLoadingScreen('Loading aircrafts');
SetLength(AirCraft, 0);
Matrix_SetIdentity(M);
SetLength(GameClass.AirCraft, 2);
Matrix_SetRotation(M, V3(0, 0, 0));
Matrix_SetTransform(M, V3(0, 250, 0));
Log.Add(TimeToStr(Now)+':  Loading aircraft "'+AirCraftInfo[pPlane1].Name+'" for player 1');
AirCraft[0] := TAirCraft.Create('data\planes\'+AirCraftInfo[pPlane1].Name+'\airplane.3ds', NewtonWorld, M);
with AirCraft[0] do
 begin
 Speed     := AirCraftInfo[pPlane1].Speed;
 FirePower := AirCraftInfo[pPlane1].FirePower;
 // Assign controller
 if Settings.Controller1 = 'keyboard' then
  Controller := ctKeyboard
 else
  begin
  Controller  := ctJoystick;
  JoystickID  := StrToInt(Settings.Controller1);
  SDLJoystick := SDL_JoystickOpen(JoystickID);
  end;
 SetLength(Emitter.Texture, 1);
 Emitter.Texture[0] := BindTexture('particle_smoke');
 Pause       := @self.Pause;
 MuzzleTexID := GameClass.BindTexture('muzzle');
 end;
Matrix_SetTransform(M, V3(125, 250, 0));
Matrix_SetRotation(M, V3(0, 180, 0));
Log.Add(TimeToStr(Now)+':  Aircraft for player 1 loaded');
Log.Add(TimeToStr(Now)+':  Loading aircraft "'+AirCraftInfo[pPlane2].Name+'" for player 2');
AirCraft[1] := TAirCraft.Create('data\planes\'+AirCraftInfo[pPlane2].Name+'\airplane.3ds', NewtonWorld, M);
with AirCraft[1] do
 begin
 Speed     := AirCraftInfo[pPlane2].Speed;
 FirePower := AirCraftInfo[pPlane2].FirePower;
 // Assign controller
 if Settings.Controller2 = 'keyboard' then
  Controller := ctKeyboard
 else
  begin
  Controller  := ctJoystick;
  JoystickID  := StrToInt(Settings.Controller2);
  SDLJoystick := SDL_JoystickOpen(JoystickID);
  end;
 SetLength(Emitter.Texture, 1);
 Emitter.Texture[0] := BindTexture('particle_smoke');
 Pause       := @self.Pause;
 MuzzleTexID := GameClass.BindTexture('muzzle');
 end;
Log.Add(TimeToStr(Now)+':  Aircraft for player 2 loaded');
UpdateLoadingScreen('Loading scenario');
LoadScenario('data\scenarios\'+pScenario);
GameMenu.Visible      := False;
Fade                  := 1.5;
Log.Add(TimeToStr(Now)+': New game started');
SplitScreen  := not (IsServer or IsClient);
InGame       := True;
NetworkTimer := NetworkInterval;
end;

// =============================================================================
//  TGameClass.StartServer
// =============================================================================
procedure TGameClass.EndGame;
begin
{if IsServer then
 NetworkClass.CloseServerSocket;
if IsClient then
 NetworkClass.CloseClientSocket;}
GameMenu.SetActiveSubMenu('mainmenu');
InGame           := False;
IsClient         := False;
IsServer         := False;
GameMenu.Visible := True;
end;

// =============================================================================
//  TGameClass.StartServer
// =============================================================================
procedure TGameClass.StartServer;
begin
NetworkClass.OnServerEnd := EndServer;
if not NetworkClass.OpenServerSocket(Settings.TCPPort) then
 begin
 ShowMessage('Could not start server! Please see log for more infos!');
 GameMenu.SetActiveSubMenu('mainmenu');
 exit;
 end;
WaitForClient := True;
InGame        := False;
IsClient      := False;
StartNewGame(ScenarioName[GameMenu.SelScenario], GameMenu.SelAircraft[0], GameMenu.SelAircraft[1], DC);
end;

// =============================================================================
//  TGameClass.EndServer
// =============================================================================
procedure TGameClass.EndServer;
begin
EndGame;
ShowMessage('Client has disconnected!');
end;

// =============================================================================
//  TGameClass.StartClient
// =============================================================================
procedure TGameClass.StartClient;
var
 IP : String;
begin
NetworkClass.OnClientEnd := EndClient;
IP := '192.168.0.1';
if not InputQuery('Server IP', 'IP of Server', IP) then
 exit;
IsClient      := True;
InGame        := False;
WaitForServer := True;
if not NetworkClass.OpenClientSocket(IP, Settings.TCPPort) then
 begin
 NetworkClass.CloseClientSocket;
 IsClient := False;
 ShowMessage('Could not connect! Server did not respond in given time. (Wrong IP? Server not running?)');
 end;
end;

// =============================================================================
//  TGameClass.EndClient
// =============================================================================
procedure TGameClass.EndClient;
begin
EndGame;
ShowMessage('Disconnected from server (server closed the game)!');
end;

// =============================================================================
//  TGameClass.CheckNetwork
// =============================================================================
procedure TGameClass.CheckNetwork;
var
 Plane1, Plane2 : Integer;
 i              : Integer;
begin
// Client to server messages
if IsServer then
 case NetworkClass.LastClientPackage.MessageType of
  nmClient_ConnectionACK :
   begin
   NetworkClass.ServerSendPackage(nmServer_RandomSeed, IntToStr(NetworkClass.ServerRandSeed), '', '');
   NetworkSleep(125);
   NetworkClass.ServerSendPackage(nmServer_ScenarioInfo, ScenarioName[GameMenu.SelScenario], AirCraftInfo[GameMenu.SelAircraft[0]].Name, AirCraftInfo[GameMenu.SelAircraft[1]].Name);
   NetworkSleep(125);
   end;
  nmClient_ScenarioInfoACK :
   begin
//   StartNewGame(ScenarioName[GameMenu.SelScenario], GameMenu.SelAircraft[0], GameMenu.SelAircraft[1], DC);
   NetworkClass.ServerSendPackage(nmServer_IsReady, '', '', '');
   end;
  nmClient_IsReady :
   begin
   WaitForClient := False;
   end;
  nmClient_InputInfo :
   if Length(AirCraft) > 0 then
    if Assigned(AirCraft[1]) then
     begin
     with NetworkClass.LastClientPackage.InputInfo do
      begin
      AirCraft[1].RollLeft    := RollLeft;
      AirCraft[1].RollRight   := RollRight;
      AirCraft[1].PitchFront  := PitchFront;
      AirCraft[1].PitchBack   := PitchBack;
      AirCraft[1].PaddleLeft  := PaddleLeft;
      AirCraft[1].PaddleRight := PaddleRight;
      end;
     with NetworkClass.LastClientPackage.AirCraftInfo[0] do
      begin
      AirCraft[1].IsFiring    := IsFiring;
      AirCraft[1].Throttle    := Throttle;
      end;
     end;
 end;
// Server to client messages
if IsClient then
 case NetworkClass.LastServerPackage.MessageType of
  nmServer_ScenarioInfo :
   begin
   Plane1 := 0;
   Plane2 := 0;
   for i := 0 to High(AirCraftInfo) do
    begin
    if AirCraftInfo[i].Name = NetworkClass.LastServerPackage.StringC then
     Plane1 := i;
    if AirCraftInfo[i].Name = NetworkClass.LastServerPackage.StringB then
     Plane2 := i;
    end;
   StartNewGame(NetworkClass.LastServerPackage.StringA, Plane1, Plane2, DC);
   NetworkClass.ClientSendPackage(nmClient_IsReady, '', '', '');
   IsClient             := True;
   AirCraft[0].IsRemote := True;
   AirCraft[0].IsClient := True;
   AirCraft[1].IsRemote := True;
   end;
  nmServer_RandomSeed :
   begin
   NetworkClass.ServerRandSeed := StrToInt(NetworkClass.LastServerPackage.StringA);
   RandSeed := NetworkClass.ServerRandSeed;
   NetworkClass.Log.Add('TGameClass.CheckNetwork : Received random seed ('+IntToStr(RandSeed)+') from server');
   end;
  nmServer_IsReady :
   begin
   WaitForServer := False;
   end;
  nmServer_AirCraftMatrix :
   if Length(AirCraft) > 0 then
    begin
    if Assigned(AirCraft[0]) then
     begin
     AirCraft[0].Matrix         := NetworkClass.LastServerPackage.Matrix;
     AirCraft[0].Health         := NetworkClass.LastServerPackage.AirCraftInfo[1].Health;
     AirCraft[0].GunHeat        := NetworkClass.LastServerPackage.AirCraftInfo[1].GunHeat;
     AirCraft[0].Lifes          := NetworkClass.LastServerPackage.AirCraftInfo[1].Lifes;
     NewtonBodySetMatrix(AirCraft[0].NewtonBody, @AirCraft[0].Matrix[0,0]);
     end;
    if Assigned(AirCraft[1]) then
     begin
     AirCraft[1].Matrix         := NetworkClass.LastServerPackage.MatrixB;
     AirCraft[1].Health         := NetworkClass.LastServerPackage.AirCraftInfo[0].Health;
     AirCraft[1].GunHeat        := NetworkClass.LastServerPackage.AirCraftInfo[0].GunHeat;
     AirCraft[1].Throttle       := NetworkClass.LastServerPackage.AirCraftInfo[0].Throttle;
     AirCraft[1].Lifes          := NetworkClass.LastServerPackage.AirCraftInfo[0].Lifes;
     AirCraft[1].IsFiring       := NetworkClass.LastServerPackage.AirCraftInfo[0].IsFiring;
     NewtonBodySetMatrix(AirCraft[1].NewtonBody, @AirCraft[1].Matrix[0,0]);
     end;
    end;
  nmServer_BonusCrateDrop :
   begin
   SpawnBonusCrate;
   BonusCrate[High(BonusCrate)].Bonus := StrToInt(NetworkClass.LastServerPackage.StringA);
   NewtonBodySetMatrix(BonusCrate[High(BonusCrate)].NewtonBody, @NetworkClass.LastServerPackage.Matrix[0,0]);
   end;
 end;
NetworkClass.LastClientPackage.MessageType := 0;
NetworkClass.LastServerPackage.MessageType := 0;
// Update aircraft info
if InGame then
 begin
 NetworkTimer := NetworkTimer - TimeFactor;
 if NetworkTimer <= 0 then
  begin
  NetworkTimer := NetworkInterval;
  if IsServer and (not WaitForClient) and (Length(AirCraft) > 0) then
   if Assigned(AirCraft[1]) then
    NetworkClass.ServerSendAircraftMatrix(TimeFactor, AirCraft[1], AirCraft[0]);
  if (IsClient) and (Length(AirCraft) > 0) then
   if Assigned(AirCraft[0]) then
    NetworkClass.ClientSendAircraftInput(TimeFactor, AirCraft[0]);
  end;
 end;
end;

// =============================================================================
//  TGameClass.SpawnBonusCrate
// =============================================================================
procedure TGameClass.SpawnBonusCrate;
var
 TmpV     : TVector3f;
 TmpM     : TMatrix4f;
 Collider : PNewtonCollision;
 Mass     : Single;
 Ixx      : Single;
 Iyy      : Single;
 Izz      : Single;
 vMin     : TVector3f;
 vMax     : TVector3f;
 BoxSize  : TVector3f;
 ExitLoop : Boolean;
begin
Collider := NewtonCreateConvexHull(NewtonWorld, CrateModel.Mesh[0].NumVertex, @CrateModel.Mesh[0].FVertex[0], SizeOf(TVector3f), nil);
SetLength(BonusCrate, Length(BonusCrate)+1);
BonusCrate[High(BonusCrate)] := TBonusCrate.Create;
with BonusCrate[High(BonusCrate)] do
 begin
 // Get a random position that is at least 100 units away from the aircrafts
 repeat
 ExitLoop := True;
 TmpV     := V3((Random-Random) * Terrain.Width/2 * 0.75, 600, (Random-Random) * Terrain.Height/2 * 0.75);
 if (VDistance(AirCraft[0].GetPosition, TmpV) < 100) or (VDistance(AirCraft[1].GetPosition, TmpV) < 100) then
  ExitLoop := False;
 until ExitLoop;
 Matrix_SetIdentity(TmpM);
 Matrix_SetTransform(TmpM, TmpV);
 NewtonBody := NewtonCreateBody(NewtonWorld, Collider);
 Mass    := 5;
 NewtonBodyGetAABB(NewtonBody, @vMin[0], @vMax[0]);
 BoxSize := V3(Abs(vMax[0]-vMin[0]), Abs(vMax[1]-vMin[1]), Abs(vMax[2]-vMin[2]));
 Ixx := Mass * (BoxSize[1] * BoxSize[1] + BoxSize[2] * BoxSize[2]) / 12;
 Iyy := Mass * (BoxSize[0] * BoxSize[0] + BoxSize[2] * BoxSize[2]) / 12;
 Izz := Mass * (BoxSize[0] * BoxSize[0] + BoxSize[1] * BoxSize[1]) / 12;
 NewtonBodySetMassMatrix(NewtonBody, Mass, Ixx, Iyy, Izz);
 NewtonBodySetMatrix(NewtonBody, @TmpM[0,0]);
 NewtonBodySetForceAndTorqueCallBack(NewtonBody, PhysicsApplyGravityForce);
 NewtonReleaseCollision(NewtonWorld, Collider);
 // The plane should never get into rest state
 NewtonBodySetAutoFreeze(NewtonBody, 0);
 NewtonWorldUnfreezeBody(NewtonWorld, NewtonBody);
 // Assign this plane with the newton body
 NewtonBodySetUserData(NewtonBody, self);
 NewtonBodySetMaterialGroupID(NewtonBody, MatBonusCrate);
 SoundSystem.PlaySample('cratedrop', False);
 Bonus := Random(4);
 if IsServer then
  NetworkClass.ServerSendBonusCrateDrop(TmpV, Bonus);
 end;
end;


// *****************************************************************************
//  TCloudLayer
// *****************************************************************************

// =============================================================================
//  TCloudLayer.Create
// =============================================================================
procedure TCloudLayer.Generate(pNumberOfClouds : Integer);
var
 i : Integer;
begin
SetLength(Cloud, pNumberOfClouds);
for i := 0 to High(Cloud) do
 Cloud[i] := V3(Random(50000)/10-Random(50000)/10, 1000+Random(300)-Random(300), Random(50000)/10-Random(50000)/10);
end;

// =============================================================================
//  TCloudLayer.Create
// =============================================================================
procedure TCloudLayer.AddCloudTexture(pTextureFile : String);
begin
SetLength(CloudTexture, Length(CloudTexture)+1);
LoadTexture(pTextureFile, CloudTexture[High(CloudTexture)]);
end;

// =============================================================================
//  TCloudLayer.Render
// =============================================================================
procedure TCloudLayer.Render(pViewerPos : TVector3f);
var
 j      : Integer;
 TmpM   : TMatrix4f;
 TmpV   : TVector3f;
begin
CloudsDrawn := 0;
glGetFloatv(GL_MODELVIEW_MATRIX, @TmpM[0,0]);
glPushMatrix;
 glDisable(GL_CULL_FACE);
 glEnable(GL_TEXTURE_2D);
 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 glDepthMask(False);
 glBindTexture(GL_TEXTURE_2D, CloudTexture[0]);
 glLoadIdentity;
 glBegin(GL_QUADS);
  for j := 0 to High(Cloud) do
   if Abs(VDistance(pViewerPos, Cloud[j])) < zFar then
    if Frustum.IsSphereWithin(Cloud[j][0], Cloud[j][1], Cloud[j][2], 240) then
     begin
     // Fade out after 1/2 of zFar
     if Abs(VDistance(pViewerPos, Cloud[j])) > zFar/2 then
      glColor4f(1,1,1, (zFar-Abs(VDistance(pViewerPos, Cloud[j])))/(zFar/2)*0.4)
     else
      glColor4f(1,1,1,0.4);
     TmpV := Matrix_TansformVector(TmpM, Cloud[j]);
     glTexCoord2f(1,1); glVertex3f(TmpV[0]-240, TmpV[1]-200, TmpV[2]);
     glTexCoord2f(1,0); glVertex3f(TmpV[0]-240, TmpV[1]+200, TmpV[2]);
     glTexCoord2f(0,0); glVertex3f(TmpV[0]+240, TmpV[1]+200, TmpV[2]);
     glTexCoord2f(0,1); glVertex3f(TmpV[0]+240, TmpV[1]-200, TmpV[2]);
     inc(CloudsDrawn);
     end;
 glEnd;
glPopMatrix;
glColor3f(1,1,1);
glDepthMask(True);
glDisable(GL_BLEND);
end;

// =============================================================================
//  TCloudLayer.Destroy
// =============================================================================
destructor TCloudLayer.Destroy;
var
 i : Integer;
begin
if Length(CloudTexture) > 0 then
 for i := 0 to High(CloudTexture) do
  glDeleteTextures(1, @CloudTexture[i]);
SetLength(Cloud, 0);
SetLength(CloudTexture, 0);
end;


// *****************************************************************************
//  TBonusCrate
// *****************************************************************************

// =============================================================================
//  TBonusCrate.Render
// =============================================================================
procedure TBonusCrate.Render;
const
 LightPos     : array[0..3] of Single = (0, -10, 0, 1);
 LightAmbient : array[0..3] of Single = (0.5, 0.5, 0.5, 1);
begin
glDisable(GL_CULL_FACE);
NewtonBodyGetMatrix(NewtonBody, @Matrix[0,0]);
glPushMatrix;
 glMultMatrixf(@Matrix[0,0]);
 glLightfv(GL_LIGHT0, GL_POSITION, @LightPos[0]);
 glLightfv(GL_LIGHT0, GL_AMBIENT,  @LightAmbient[0]);
 glEnable(GL_LIGHTING);
 glEnable(GL_LIGHT0);
 GameClass.CrateModel.Render;
 glDisable(GL_LIGHTING);
glPopMatrix;
glEnable(GL_CULL_FACE);
end;

// =============================================================================
//  TBonusCrate.Apply
// =============================================================================
procedure TBonusCrate.Apply(pAirCraft : TAirCraft);
begin
if Remove then
 exit;
case Bonus of
 0 : begin // Life
     if pAirCraft.Lifes < MaxLifes then
      inc(pAirCraft.Lifes);
     end;
 1 : begin // Faster gun cooling
     pAirCraft.GunCooling := pAirCraft.GunCooling + 0.5;
     end;
 2 : begin // Repair
     pAirCraft.Health := 100;
     end;
 3 : begin // Speed
     pAirCraft.SpeedModifier := pAirCraft.SpeedModifier + 0.5;
     end;
end;
pAirCraft.LastBonus      := Bonus;
pAirCraft.LastBonusTimer := 10;
SoundSystem.PlaySample('bonuspickup', False);
end;

// =============================================================================
//  TBonusCrate.Destroy
// =============================================================================
destructor TBonusCrate.Destroy;
begin
NewtonDestroyBody(GameClass.NewtonWorld, NewtonBody);
end;

initialization
 Log := TStringList.Create;

finalization
 Log.SaveToFile(ExtractFilePath(Application.ExeName)+'log.txt');

end.
