// =============================================================================
//
//  JagdGeschwader_Network.pas
//
// =============================================================================
//
//   Copyright  2005 by Sascha Willems (www.delphigl.de)
//
// =============================================================================
//
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
//
// =============================================================================
//
//   Entry for the 2005 PGD "Dog-Fight" Competition
//
// =============================================================================

unit JagdGeschwader_Network;

interface

uses
 SysUtils,
 Forms,
 ScktComp,
 Controls,
 Classes,
 Dialogs,

 dglOpenGL,
 glMatrixHelper,
 NewtonImport,

 JagdGeschwader_AirCraft;

const
 // Networkmessages
 // Server
 nmServer_ConnectionAccepted = $01;
 nmServer_GameIsFull         = $02;
 nmServer_ScenarioInfo       = $03;
 nmServer_ClosingServer      = $04;
 nmServer_AircraftMatrix     = $05;
 nmServer_IsReady            = $06;
 nmServer_BonusCrateDrop     = $07;
 nmServer_RemoveBonusCrate   = $08;
 nmServer_RandomSeed         = $09;
 // Client
 nmClient_ConnectionACK      = $01;
 nmClient_ScenarioInfoACK    = $02;
 nmClient_InputInfo          = $04;
 nmClient_IsReady            = $05;

type
 TServerEvent = procedure of object;
 TClientEvent = procedure of object;
 // TInputInfo =================================================================
 TInputInfo = record
   RollLeft    : Boolean;
   RollRight   : Boolean;
   PitchFront  : Boolean;
   PitchBack   : Boolean;
   PaddleLeft  : Boolean;
   PaddleRight : Boolean;
  end;
 // TAirCraftInfo =============================================================
 TAirCraftInfo = record
   IsFiring : Boolean;
   Throttle : Single;
   GunHeat  : Single;
   Health   : Single;
   Lifes    : Byte;
  end;
 // TNetworkInfoPackage ========================================================
 TNetworkInfoPackage = record
   MessageType  : Byte;
   StringA      : String[25];
   StringB      : String[25];
   StringC      : String[25];
   Matrix       : TMatrix4f;                    // From Server : Matrix of client's plane
   MatrixB      : TMatrix4f;                    // From Server : Matrix of server's plane
   InputInfo    : TInputInfo;                   // From Client
   AirCraftInfo : array[0..1] of TAirCraftInfo; // From Server
  end;
 // TNetworkClass ==============================================================
 TNetworkClass = class
   ServerSocket      : TServerSocket;
   ClientSocket      : TClientSocket;
   LastError         : String;
   Log               : TStringList;
   LastClientPackage : TNetworkInfoPackage;
   LastServerPackage : TNetworkInfoPackage;
   ServerRandSeed    : LongInt;
   // Callbacks
   OnServerEnd       : TServerEvent;
   OnClientEnd       : TClientEvent;
   // Server
   function OpenServerSocket(pPort : Word) : Boolean;
   procedure CloseServerSocket;
   procedure ServerSendPackage(pPackageID : Byte; pStringA, pStringB, pStringC : String);
   procedure ServerSendAircraftMatrix(pTimeFactor : Single; pAirCraftA, pAirCraftB : TAirCraft);
   procedure ServerSendBonusCrateDrop(pPosition : TVector3f; pBonus : Byte);
   // Server events
   procedure OnClientConnect(Sender: TObject; Socket: TCustomWinSocket);
   procedure OnClientDisconnect(Sender: TObject; Socket: TCustomWinSocket);
   procedure OnClientRead(Sender: TObject; Socket: TCustomWinSocket);
   procedure OnClientError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
   // Client
   function OpenClientSocket(pHost : String; pPort : Word) : Boolean;
   procedure CloseClientSocket;
   procedure OnServerDisconnect(Sender: TObject; Socket: TCustomWinSocket);
   procedure ClientSendPackage(pPackageID : Byte; pStringA, pStringB, pStringC : String);
   procedure ClientSendAircraftInput(pTimeFactor : Single; pAirCraft : TAirCraft);
   // Client events
   procedure OnServerRead(Sender: TObject; Socket: TCustomWinSocket);
   procedure OnServerError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
  end;

var
 NetworkClass  : TNetworkClass;

procedure NetworkSleep(pMS : Integer);

implementation

// =============================================================================
//  NetworkSleep
// =============================================================================
procedure NetworkSleep(pMS : Integer);
var
 i : Integer;
begin
for i := 0 to (pMS div 5)-1 do
 begin
 Application.ProcessMessages;
 Sleep(5);
 end;
end;

// *****************************************************************************
//  TNetworkClass
// *****************************************************************************

// =============================================================================
//  TNetworkClass.OpenServerSocket
// =============================================================================
function TNetworkClass.OpenServerSocket(pPort : Word) : Boolean;
begin
if not Assigned(ServerSocket) then
 begin
 ServerSocket                    := TServerSocket.Create(nil);
 ServerSocket.OnClientConnect    := OnClientConnect;
 ServerSocket.OnClientDisconnect := OnClientDisconnect;
 ServerSocket.OnClientRead       := OnClientRead;
 ServerSocket.OnClientError      := OnClientError;
 end;
try
 if ServerSocket.Active then
  begin
  Log.Add('TNetworkClass.OpenServerSocket : Server was already active. Closed before restarting.');
  ServerSocket.Close;
  end;
 ServerSocket.Port := pPort;
 ServerSocket.Open;
 Log.Add('TNetworkClass.OpenServerSocket : Server started at Port '+IntToStr(pPort));
 Result := True;
except
 ServerSocket.Free;
 Log.Add('NetworkClass.OpenServerSocket : Could not start Server!');
 Result := False;
end;
end;

// =============================================================================
//  TNetworkClass.CloseServerSocket
// =============================================================================
procedure TNetworkClass.CloseServerSocket;
begin
if Assigned(ServerSocket) then
 begin
 ServerSocket.Close;
 //ServerSocket.Free;
 Log.Add('TNetworkClass.CloseServerSocket : Server closed.');
 end;
end;

// =============================================================================
//  TNetworkClass.ServerSendPackage
// =============================================================================
procedure TNetworkClass.ServerSendPackage(pPackageID : Byte; pStringA, pStringB, pStringC : String);
var
 NetPackage : TNetworkInfoPackage;
begin
NetPackage.MessageType := pPackageID;
NetPackage.StringA     := pStringA;
NetPackage.StringB     := pStringB;
NetPackage.StringC     := pStringC;
if ServerSocket.Socket.ActiveConnections > 0 then
 if ServerSocket.Socket.Connections[0].Connected then
  if ServerSocket.Socket.Connections[0].SendBuf(NetPackage, SizeOf(NetPackage)) = SizeOf(NetPackage) then
   Log.Add('TNetworkClass.ServerSendPackage : Info package with ID $'+IntToHex(pPackageID, 2)+' send to client')
  else
   Log.Add('TNetworkClass.ServerSendPackage : Could not send package with ID $'+IntToHex(pPackageID, 2));
end;

// =============================================================================
//  TNetworkClass.ServerSendAircraftMatrix
// =============================================================================
procedure TNetworkClass.ServerSendAircraftMatrix(pTimeFactor : Single; pAirCraftA, pAirCraftB : TAirCraft);
var
 NetPackage : TNetworkInfoPackage;
begin
if not Assigned(ServerSocket) then
 exit;
if not ServerSocket.Active then
 exit;
NetPackage.Matrix      := pAirCraftA.Matrix;
NetPackage.MatrixB     := pAirCraftB.Matrix;
NetPackage.MessageType := nmServer_AircraftMatrix;
// Server plane
with NetPackage.AirCraftInfo[0] do
 begin
 Health   := pAirCraftB.Health;
 Lifes    := pAirCraftB.Lifes;
 GunHeat  := pAirCraftB.GunHeat;
 Throttle := pAirCraftB.Throttle;
 IsFiring := pAirCraftB.IsFiring;
 end;
// Client plane
with NetPackage.AirCraftInfo[1] do
 begin
 Health   := pAirCraftA.Health;
 Lifes    := pAirCraftA.Lifes;
 GunHeat  := pAirCraftA.GunHeat;
 Throttle := pAirCraftA.Throttle;
 end;
if ServerSocket.Socket.ActiveConnections > 0 then
 if ServerSocket.Socket.Connections[0].Connected then
  ServerSocket.Socket.Connections[0].SendBuf(NetPackage, SizeOf(NetPackage));
end;

// =============================================================================
//  TNetworkClass.ServerSendBonusCrateDrop
// =============================================================================
procedure TNetworkClass.ServerSendBonusCrateDrop(pPosition : TVector3f; pBonus : Byte);
var
 NetPackage : TNetworkInfoPackage;
begin
NetPackage.MessageType := nmServer_BonusCrateDrop;
Matrix_SetIdentity(NetPackage.Matrix);
Matrix_SetTransform(NetPackage.Matrix, pPosition);
NetPackage.StringA := IntToStr(pBonus);
if ServerSocket.Socket.ActiveConnections > 0 then
 if ServerSocket.Socket.Connections[0].Connected then
  if ServerSocket.Socket.Connections[0].SendBuf(NetPackage, SizeOf(NetPackage)) = SizeOf(NetPackage) then
   Log.Add('TNetworkClass.ServerSendBonusCrateDrop : Package send to client')
  else
   Log.Add('TNetworkClass.ServerSendBonusCrateDrop : Could not send package');
end;

// =============================================================================
//  TNetworkClass.OnClientConnect
// =============================================================================
procedure TNetworkClass.OnClientConnect(Sender: TObject; Socket: TCustomWinSocket);
begin
if ServerSocket.Socket.ActiveConnections > 1 then
 begin
 Log.Add('TServer.OnClientConnect : Client @ '+Socket.RemoteAddress+'refused. Game is full!');
 ServerSendPackage(nmServer_GameIsFull, '', '', '');
 exit;
 end;
Log.Add('TServer.OnClientConnect : New client @ '+Socket.RemoteAddress);
ServerSendPackage(nmServer_ConnectionAccepted, '', '', '');
end;

// =============================================================================
//  TNetworkClass.OnClientDisconnect
// =============================================================================
procedure TNetworkClass.OnClientDisconnect(Sender: TObject; Socket: TCustomWinSocket);
begin
Log.Add('TServer.OnClientDisconnect : Client @ '+Socket.RemoteAddress+' has disconnected from the server. Closing server!');
if Assigned(OnServerEnd) then
 OnServerEnd;
end;

// =============================================================================
//  TNetworkClass.OnClientRead
// =============================================================================
procedure TNetworkClass.OnClientRead(Sender: TObject; Socket: TCustomWinSocket);
var
 NetPackage : TNetworkInfoPackage;
begin
Socket.ReceiveBuf(NetPackage, SizeOf(NetPackage));
case NetPackage.MessageType of
 nmClient_ConnectionACK :
  Log.Add('TNetworkClass.OnClientRead : Recieved nmClient_ConnectionACK from client @ '+Socket.RemoteAddress);
 nmClient_ScenarioInfoACK :
  Log.Add('TNetworkClass.OnClientRead : Recieved nmClient_ScenarioInfoACK from client @ '+Socket.RemoteAddress);
 nmClient_IsReady :
  Log.Add('TNetworkClass.OnClientRead : Recieved nmClient_IsReady from client @ '+Socket.RemoteAddress);
end;
LastClientPackage := NetPackage;
end;

// =============================================================================
//  TNetworkClass.OnClientError
// =============================================================================
procedure TNetworkClass.OnClientError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
begin
case ErrorEvent of
 eeGeneral    :	Log.Add('TNetworkClass.OnClientError : General socket error');
 eeSend	      : Log.Add('TNetworkClass.OnClientError : Could not write to socket connection');
 eeReceive    :	Log.Add('TNetworkClass.OnClientError : Could not read from socket connection');
 eeConnect    :	Log.Add('TNetworkClass.OnClientError : Could not finish connection request');
 eeDisconnect	: Log.Add('TNetworkClass.OnClientError : Error on closing connection');
 eeAccept	    : Log.Add('TNetworkClass.OnClientError : Could not accept client connection request');
end;
ErrorCode := 0;
if Assigned(OnServerEnd) then
 OnServerEnd;
end;

// =============================================================================
//  TNetworkClass.OpenClientSocket
// =============================================================================
function TNetworkClass.OpenClientSocket(pHost : String; pPort : Word) : Boolean;
begin
Result := False;
if not Assigned(ClientSocket) then
 begin
 ClientSocket              := TClientSocket.Create(nil);
 ClientSocket.OnRead       := OnServerRead;
 ClientSocket.OnDisconnect := OnServerDisconnect;
 ClientSocket.OnError      := OnServerError;
 end;
if ClientSocket.Active and  ClientSocket.Socket.Connected then
 if MessageDlg('You are already connected to a server (IP:'+ClientSocket.Socket.RemoteAddress+')!'+#13+'Really disconnect?', mtConfirmation, [mbYes, mbNo], 0) = mrYes then
  begin
  Log.Add('TNetworkClass.OpenClientSocket : Connection to server @ '+ClientSocket.Socket.RemoteAddress+' closed');
  ClientSocket.Close;
  end
 else
  exit;
try
 if ClientSocket.Active then
  ClientSocket.Close;
 ClientSocket.Port := pPort;
 ClientSocket.Host := pHost;
 Log.Add('TNetworkClass.OpenClientSocket : Client started at port '+IntToStr(pPort));
 ClientSocket.Open;
except
 ClientSocket.Free;
 Log.Add('TNetworkClass.OpenClientSocket : Client could not be started!!');
end;
NetworkSleep(2000);
Result := ClientSocket.Active;
end;

// =============================================================================
//  TNetworkClass.CloseClientSocket
// =============================================================================
procedure TNetworkClass.CloseClientSocket;
begin
if Assigned(ClientSocket) then
 begin
 ClientSocket.Close;
 //ClientSocket.Free;
 Log.Add('TNetworkClass.CloseClientSocket : Client closed.');
 end;
end;

// =============================================================================
//  TNetworkClass.OnServerDisconnect
// =============================================================================
procedure TNetworkClass.OnServerDisconnect(Sender: TObject; Socket: TCustomWinSocket);
begin
Log.Add('TNetworkClass.OnServerDisconnect : Server @ '+Socket.RemoteAddress+' was closed, or connection was lost!');
if Assigned(OnClientEnd) then
 OnClientEnd;
end;

// =============================================================================
//  TNetworkClass.OnServerError
// =============================================================================
procedure TNetworkClass.OnServerError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
begin
case ErrorEvent of
 eeGeneral    :	Log.Add('TNetworkClass.OnClientError : General socket error');
 eeSend	      : Log.Add('TNetworkClass.OnClientError : Could not write to socket connection');
 eeReceive    :	Log.Add('TNetworkClass.OnClientError : Could not read from socket connection');
 eeConnect    :	Log.Add('TNetworkClass.OnClientError : Could not finish connection request');
 eeDisconnect	: Log.Add('TNetworkClass.OnClientError : Error on closing connection');
 eeAccept	    : Log.Add('TNetworkClass.OnClientError : Could not accept client connection request');
end;
ErrorCode := 0;
if Assigned(OnClientEnd) then
 OnClientEnd;
end;

// =============================================================================
//  TNetworkClass.ClientSendPackage
// =============================================================================
procedure TNetworkClass.ClientSendPackage(pPackageID : Byte; pStringA, pStringB, pStringC : String);
var
 NetPackage : TNetworkInfoPackage;
begin
if not Assigned(ClientSocket) then
 exit;
NetPackage.MessageType := pPackageID;
NetPackage.StringA     := pStringA;
NetPackage.StringB     := pStringB;
NetPackage.StringC     := pStringC;
if ClientSocket.Active then
 begin
 if ClientSocket.Socket.Connected then
  if ClientSocket.Socket.SendBuf(NetPackage, SizeOf(NetPackage)) = SizeOf(NetPackage) then
   Log.Add('TNetworkClass.ClientSendPackage : Info package with ID $'+IntToHex(pPackageID, 2)+' send to server')
  else
   Log.Add('TNetworkClass.ClientSendPackage : Could not send package with ID $'+IntToHex(pPackageID, 2));
 end
else
 Log.Add('TNetworkClass.ClientSendPackage : Could not send package with ID $'+IntToHex(pPackageID, 2)+' cause client socket is not active!');
end;

// =============================================================================
//  TNetworkClass.ClientSendAircraftInput
// =============================================================================
procedure TNetworkClass.ClientSendAircraftInput(pTimeFactor : Single; pAirCraft : TAirCraft);
var
 NetPackage : TNetworkInfoPackage;
begin
if not Assigned(ClientSocket) then
 exit;
if not ClientSocket.Active then
 exit;
NetPackage.MessageType := nmClient_InputInfo;
with NetPackage.InputInfo do
 begin
 RollLeft    := pAirCraft.RollLeft;
 RollRight   := pAirCraft.RollRight;
 PitchFront  := pAirCraft.PitchFront;
 PitchBack   := pAirCraft.PitchBack;
 PaddleLeft  := pAirCraft.PaddleLeft;
 PaddleRight := pAirCraft.PaddleRight;
 end;
with NetPackage.AirCraftInfo[0] do
 begin
 IsFiring   := pAirCraft.IsFiring;
 Throttle   := pAirCraft.Throttle;
 end;
if ClientSocket.Socket.Connected then
 ClientSocket.Socket.SendBuf(NetPackage, SizeOf(NetPackage));
//Log.Add('TNetworkClass.ClientSendPackage : Input info send to server');
end;

// =============================================================================
//  TNetworkClass.OnServerRead
// =============================================================================
procedure TNetworkClass.OnServerRead(Sender: TObject; Socket: TCustomWinSocket);
var
 NetPackage : TNetworkInfoPackage;
begin
Socket.ReceiveBuf(NetPackage, SizeOf(NetPackage));
case NetPackage.MessageType of
 nmServer_ConnectionAccepted :
  ClientSendPackage(nmClient_ConnectionACK, '', '', '');
 nmServer_ScenarioInfo :
  ClientSendPackage(nmClient_ScenarioInfoACK, '', '', '');
end;
LastServerPackage := NetPackage;
//Log.Add('TNetworkClass.OnServerRead : Recieved package with ID $'+IntToHex(NetPackage.MessageType, 2));
end;

initialization
 NetworkClass     := TNetworkClass.Create;
 NetworkClass.Log := TStringList.Create;

finalization
 with NetworkClass do
  begin
  if Assigned(ClientSocket) then
   begin
   ClientSocket.Close;
   ClientSocket.Free;
   end;
  if Assigned(ServerSocket) then
   begin
   ServerSocket.Close;
   ServerSocket.Free;
   end;
  Log.SaveToFile(ExtractFilePath(Application.ExeName)+'newtork_log.txt');
  Log.Free;
  Free;
  end;

end.
