// =============================================================================
//
//  JagdGeschwader_NewtonCallbacks.pas
//   Contains some newton callbacks (like debug) that can be used globally.
//
// =============================================================================
//
//   Copyright  2005 by Sascha Willems (www.delphigl.de)
//
// =============================================================================
//
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
//
// =============================================================================

unit JagdGeschwader_NewtonCallbacks;

interface

uses
 SysUtils,
 dglOpenGL,
 glMatrixHelper,
 Dialogs,
 NewtonImport;

var
 PickedParam   : Single;      // Needed for RayCastFilter
 PickedBody    : PNewtonBody; // Ditto
 PickedIgnore  : PNewtonBody; // Ditto, tells the callback what body to ignore (no self hitting)
 // Materials for collision detection
 MatTerrain    : Integer;     // Used in collision to find out what collided
 MatAirCraft   : Integer;     // Used in collision to find out what collided
 MatBonusCrate : Integer;     // Used in collision to find out what collided

procedure Debug_ShowBodyCollision(const Body : PNewtonBody); cdecl;
procedure Debug_ShowGeometryCollision(const Body : PNewtonBody; VertexCount : Integer; const FaceArray : PFloat; FaceId : int); cdecl;
procedure PhysicsApplyGravityForce(const Body : PNewtonBody); cdecl;
function  RayCastFilter(const Body : PNewtonBody; const Normal : PFloat; CollisionID : Int; UserData : Pointer; IntersectParam : Single) : Single; cdecl;
function GetHeight(nWorld : PNewtonWorld; pPosition : TVector3f) : Single;

implementation

// =============================================================================
//  GetHeight
// =============================================================================
//  Place an object at a given height above the terrain
// =============================================================================
function GetHeight(nWorld : PNewtonWorld; pPosition : TVector3f) : Single;
var
 LineStart : TVector3f;
 LineEnd   : TVector3f;
 HitPos    : TVector3f;
begin
LineStart    := pPosition;
LineStart[1] := LineStart[1]-2000;
LineEnd      := pPosition;
PickedBody   := nil;
PickedParam  := 1.1;
PickedIgnore := nil;
NewtonWorldRayCast(nWorld, @LineStart[0], @LineEnd[0], @RayCastFilter, nil);
HitPos    := VAdd(LineStart, VScale(VSub(LineEnd, LineStart), PickedParam));
Result    := HitPos[1];
end;

// =============================================================================
//  Debug_ShowGeometryCollision
// =============================================================================
//  Callback that shows collision geometry of a body
// =============================================================================
procedure Debug_ShowGeometryCollision(const Body : PNewtonBody; VertexCount : Integer; const FaceArray : PFloat; FaceId : int); cdecl;
var
 i     : Integer;
 v0,v1 : array[0..2] of Single;
 vA    : array of Single;
begin
if VertexCount = 0 then
 exit;
SetLength(vA, VertexCount*3);
Move(FaceArray^, vA[0], VertexCount*3*SizeOf(Single));
v0[0] := vA[(VertexCount-1)*3];
v0[1] := vA[(VertexCount-1)*3+1];
v0[2] := vA[(VertexCount-1)*3+2];
for i := 0 to VertexCount-1 do
 begin
 v1[0] := vA[i*3];
 v1[1] := vA[i*3+1];
 v1[2] := vA[i*3+2];
 glVertex3f(v0[0], v0[1], v0[2]);
 glVertex3f(v1[0], v1[1], v1[2]);
 v0 := v1;
 end;
end;

// =============================================================================
//  Debug_ShowBodyCollision
// =============================================================================
//  Show collision geometry for all bodies in the current scene
// =============================================================================
procedure Debug_ShowBodyCollision(const Body : PNewtonBody); cdecl;
begin
NewtonBodyForEachPolygonDo(Body, Debug_ShowGeometryCollision);
end;

// =============================================================================
//  RayCastFilter
// =============================================================================
//  Callback that get's the rigid body hit by a ray.
// =============================================================================
function RayCastFilter(const Body : PNewtonBody; const Normal : PFloat; CollisionID : Int; UserData : Pointer; IntersectParam : Single) : Single; cdecl;
begin
Result := PickedParam;
if Body = PickedIgnore then
 exit;
if (IntersectParam < PickedParam) then
 begin
 PickedParam := IntersectParam;
 PickedBody  := PNewtonBody(Body);
 end;
Result    := IntersectParam;
end;

// =============================================================================
//  PhysicsApplyGravityForce
// =============================================================================
//  Callback that applies gravity force to a body  (only used for falling parachutes)
// =============================================================================
procedure PhysicsApplyGravityForce(const Body : PNewtonBody); cdecl;
var
 mass  : Single;
 Ixx   : Single;
 Iyy   : Single;
 Izz   : Single;
 Force : TVector3f;
begin
NewtonBodyGetMassMatrix(Body, @mass, @Ixx, @Iyy, @Izz);
// We use a much lower gravity value instead of default 9.8 since the parachutes
// should fall slowly
Force := V3(0, Mass * -2, 0);
NewtonBodySetForce(Body, @Force[0]);
end;


end.
