// =============================================================================
//
//  JagdGeschwader_SoundSystem.pas
//   Wraps FMOD
//
// =============================================================================
//
//   Copyright  2005 by Sascha Willems (www.delphigl.de)
//
// =============================================================================
//
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
//
// =============================================================================
//
//   Entry for the 2005 PGD "Dog-Fight" Competition
//
// =============================================================================

unit JagdGeschwader_SoundSystem;

interface

uses
 Windows,
 Classes,
 SysUtils,

 fmod,
 fmoddyn,
 fmoderrors,
 fmodpresets,
 fmodtypes;

type
 // TSoundSystem ===============================================================
 TSoundSystem = class(TComponent)
  public
   SampleHandle     : array of PFSOUNDSample;
   SampleList       : TStringList;
   BackgroundStream : PFSoundStream;
   Log              : TStringList;
   Disabled         : Boolean;
   Streaming        : Boolean;
   constructor Create(AOwner : TComponent; pDisabled : Boolean); reintroduce;
   destructor Destroy; override;
   function GetSample(pSampleName : String) : PFSoundSample;
   procedure OpenBackgroundStream(pFileName : String);
   procedure PlayBackgroundStream;
   procedure StopBackgroundStream;
   procedure AddSample(pFileName, pSampleName : String);
   procedure PlaySample(pSampleName : String;pLoop : Boolean);
   procedure AddSamplesInDir(pDirName, pFileMask : String);
  end;

var
 SoundSystem : TSoundSystem;

implementation

// *****************************************************************************
//  TSoundSystem
// *****************************************************************************

// =============================================================================
//  TSoundSystem.Create
// =============================================================================
constructor TSoundSystem.Create(AOwner : TComponent; pDisabled : Boolean);
begin
inherited Create(AOwner);
if pDisabled then
 begin
 Disabled := True;
 exit;
 end;
SampleList := TStringList.Create;
FMOD_Load;
Log := TStringList.Create;
Log.Add('TfmodSoundSystem->Initializing...');
if FSOUND_Init(44100, 32, 0) then
 Log.Add('TfmodSoundSystem->FMOD initialized')
else
 begin
 Log.Add('TfmodSoundSystem->Initialized failed!');
 exit;
 end;
end;

// =============================================================================
//  TSoundSystem.Destroy
// =============================================================================
destructor TSoundSystem.Destroy;
begin
if Disabled then
 exit;
FSOUND_Stream_Stop(BackgroundStream);
FSOUND_Stream_Close(BackgroundStream);
Log.Add('TfmodSoundSystem->Shutting down...');
SampleList.Free;
FSOUND_Close;
Log.Free;
inherited;
end;

// =============================================================================
//  TSoundSystem.GetSample
// =============================================================================
function TSoundSystem.GetSample(pSampleName : String) : PFSoundSample;
var
 Index : Integer;
begin
Result := nil;
if Disabled then
 exit;
Index := SampleList.IndexOf(pSampleName);
if Index < 0 then
 halt(0);
Result := SampleHandle[Index];
end;

// =============================================================================
//  TSoundSystem.OpenBackGroundStream
// =============================================================================
procedure TSoundSystem.OpenBackGroundStream(pFileName : String);
begin
if Disabled then
 exit;
BackgroundStream := FSOUND_Stream_Open(PChar(pFileName), FSOUND_LOOP_NORMAL, 0, 0);
end;

// =============================================================================
//  TSoundSystem.PlayBackgroundStream
// =============================================================================
procedure TSoundSystem.PlayBackgroundStream;
begin
if Disabled or (BackgroundStream = nil) then
 exit;
FSOUND_Stream_Play(FSOUND_FREE, BackgroundStream);
Streaming := True;
end;

// =============================================================================
//  TSoundSystem.PlayBackgroundStream
// =============================================================================
procedure TSoundSystem.StopBackgroundStream;
begin
if Disabled or (BackgroundStream = nil) then
 exit;
FSOUND_Stream_Stop(BackgroundStream);
Streaming := False;
end;

// =============================================================================
//  TSoundSystem.AddSample
// =============================================================================
procedure TSoundSystem.AddSample(pFileName, pSampleName : String);
begin
if Disabled then
 exit;
SampleList.Add(pSampleName);
SetLength(SampleHandle, Length(SampleHandle)+1);
SampleHandle[High(SampleHandle)] := FSOUND_Sample_Load(FSOUND_FREE, PChar(pFileName), 0 , 0, 0);
end;

// =============================================================================
//  TSoundSystem.PlaySample
// =============================================================================
procedure TSoundSystem.PlaySample(pSampleName : String;pLoop : Boolean);
var
 Index   : Integer;
 Channel : Integer;
begin
if Disabled then
 exit;
Index := SampleList.IndexOf(pSampleName);
if Index < 0 then
 halt(0);
Channel := FSOUND_PlaySound(FSOUND_FREE, SampleHandle[Index]);
if pLoop then
 FSOUND_SetLoopMode(Channel, FSOUND_LOOP_NORMAL);
end;

// =============================================================================
//  TSoundSystem.AddSamplesInDir
// =============================================================================
procedure TSoundSystem.AddSamplesInDir(pDirName, pFileMask : String);
var
 SR         : TSearchRec;
 i          : Integer;
 SampleName : String;
 OldDir     : String;
begin
if Disabled then
 exit;
OldDir := GetCurrentDir;
ChDir(pDirName);
if FindFirst(pFileMask, faAnyFile, SR) = 0 then
 repeat
 SampleName := '';
 for i := 1 to Length(SR.Name) do
  begin
  if SR.Name[i] = '.' then
   break;
  SampleName := SampleName+SR.Name[i];
  end;
 AddSample(SR.Name, SampleName);
 Log.Add(SR.Name + '/' + SampleName);
 until FindNext(SR) <> 0;
FindClose(SR);
ChDir(OldDir);
end;

end.
