// =============================================================================
//
//  JagdGeschwader_Terrain.pas
//
// =============================================================================
//
//   Copyright  2005 by Sascha Willems (www.delphigl.de)
//
// =============================================================================
//
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
//
// =============================================================================
//
//   Entry for the 2005 PGD "Dog-Fight" Competition
//
// =============================================================================

unit JagdGeschwader_Terrain;

interface

uses
 SysUtils,
 Windows,
 Graphics,
 Classes,

 glOctree,
 glMatrixHelper,
 Textures,
 dglOpenGL,
 NewtonImport,

 JagdGeschwader_NewtonCallBacks;

type
 // TTerrain ===================================================================
 TTerrain = class
   VertexArray   : TVertexArray;
   HeightData    : array of array of Byte;
   BaseTexture   : glUInt;
   DetailTexture : glUInt;
   Octree        : TOctree;
   NewtonBody    : PNewtonBody;
   NewtonWorld   : PNewtonWorld;
   Width,Height  : Single;
   SizeScale     : Single;
   HeightScale   : Single;
   procedure LoadFromFile(pHeightmapFile, pBaseTextureFile, pDetailTextureFile : String; pNewtonWorld : PNewtonWorld;pSizeScale : Single = 1; pHeightScale : Single = 1; pOffset : Integer = -50; pSerializeOnly : Boolean = False);
   procedure Render(pCameraPosition : TVector3f; pMaxDrawDistance : Single);
   destructor Destroy; override;
  end;

var
 Terrain : TTerrain;  

implementation

// =============================================================================
//  TreeSerialize
// =============================================================================
procedure TreeSerialize(serializeHandle : Pointer; const buffer : Pointer; size : size_t ); cdecl;
begin
TFileStream(SerializeHandle).WriteBuffer(Buffer^, Size);
end;

// =============================================================================
//  TreeDeSerialize
// =============================================================================
procedure TreeDeSerializeprocedure( serializeHandle : Pointer; buffer : Pointer; size : size_t ); cdecl;
begin
TFileStream(SerializeHandle).ReadBuffer(Buffer^, Size);
end;

// *****************************************************************************
//  TTerrain
// *****************************************************************************

// =============================================================================
//  TTerrain.LoadFromFile
// =============================================================================
procedure TTerrain.LoadFromFile(pHeightmapFile, pBaseTextureFile, pDetailTextureFile : String; pNewtonWorld : PNewtonWorld; pSizeScale : Single = 1; pHeightScale : Single = 1; pOffset : Integer = -50; pSerializeOnly : Boolean = False);
const
 Points  : array[0..5] of array[0..1] of Integer = ((0,0),(1,0),(0,1), (1,0),(1,1),(0,1));
var
 HeightMap : TBitmap;
 x,y       : Integer;
 i,j       : Integer;
 Collider  : PNewtonCollision;
 TmpTri    : array[0..6] of TglVertex3f;
 Min,Max   : TVector3f;
 TmpM      : TMatrix4f;
 SerFile   : TFileStream;
 FromSer   : Boolean;
begin
pSizeScale   := pSizeScale * 8;
pHeightScale := pHeightScale / 75000;
HeightScale  := pHeightScale;
SizeScale    := pSizeScale;
NewtonWorld := pNewtonWorld;
if not pSerializeOnly then
 begin
 if glIsTexture(BaseTexture) then
  glDeleteTextures(1, @BaseTexture);
 if FileExists(pBaseTextureFile) then
  LoadTexture(pBaseTextureFile, BaseTexture);
 if glIsTexture(DetailTexture) then
  glDeleteTextures(1, @DetailTexture);
 if FileExists(pDetailTextureFile) then
  LoadTexture(pDetailTextureFile, DetailTexture);
 end;
HeightMap := TBitmap.Create;
HeightMap.LoadFromFile(pHeightmapFile);
i := 0;
SetLength(HeightData, HeightMap.Width, HeightMap.Height);
Collider := NewtonCreateTreeCollision(pNewtonWorld, nil);
FromSer := FileExists('serialization.dat');
if not FromSer then
 NewtonTreeCollisionBeginBuild(Collider);
for x := 0 to HeightMap.Width-2 do
 for y := 0 to HeightMap.Height-2 do
  begin
  HeightData[x,y] := HeightMap.Canvas.Pixels[x,y] div clWhite;
  SetLength(VertexArray, Length(VertexArray)+6);
  for j := 0 to High(Points) do
   begin
   VertexArray[i+j]   := glVertex(-HeightMap.Width/2*pSizeScale+(x+Points[j,0])*pSizeScale,
                                  HeightMap.Canvas.Pixels[x+Points[j,0], y+Points[j,1]]*pHeightScale+pOffset,
                                  -HeightMap.Width/2*pSizeScale+(y+Points[j,1])*pSizeScale);
   VertexArray[i+j].t := (x+Points[j,0])/HeightMap.Width;
   VertexArray[i+j].u := 1-(y+Points[j,1])/HeightMap.Height;
   TmpTri[j]          := VertexArray[i+j];
   end;
  if not FromSer then
   begin
   NewtonTreeCollisionAddFace(Collider, 3, @TmpTri[0], SizeOf(TglVertex3f), 0);
   NewtonTreeCollisionAddFace(Collider, 3, @TmpTri[3], SizeOf(TglVertex3f), 0);
   end;
  inc(i, 6);
  end;
if not FromSer then
 NewtonTreeCollisionEndBuild(Collider, 1)
else
 begin
 SerFile := TFileStream.Create('serialization.dat', fmOpenRead);
 Collider := NewtonCreateTreeCollisionFromSerialization(pNewtonWorld, nil, TreeDeSerializeprocedure, SerFile);
 SerFile.Free;
 end;
NewtonBody := NewtonCreateBody(pNewtonWorld, Collider);
Matrix_SetIdentity(TmpM);
NewtonCollisionCalculateAABB(Collider, @TmpM[0,0], @Min[0], @Max[0]);
if not FromSer then
 begin
 SerFile := TFileStream.Create('serialization.dat', fmCreate);
 NewtonTreeCollisionSerialize(Collider, TreeSerialize, SerFile);
 SerFile.Free;
 end;
NewtonReleaseCollision(pNewtonWorld, Collider);
Width  := HeightMap.Width*pSizeScale;
Height := HeightMap.Height*pSizeScale;
HeightMap.Free;
if not pSerializeOnly then
 begin
 OctreeSettings.MaxSubdivisions := 4;
 OctreeSettings.MaxTriangles    := 2048;
 OctreeSettings.MaxDrawDistance := 512000;
 Octree := TOctree.Create;
 Octree.GetSceneDimensions(VertexArray, Length(VertexArray));
 Octree.CreateNode(VertexArray, Length(VertexArray), Octree.Centre, Octree.Width);
 Octree.DrawOctree(Octree, True);
 OctreeSettings.MaxDrawDistance := 1024;
 SetLength(VertexArray, 0);
 NewtonBodySetMaterialGroupID(NewtonBody, MatTerrain);
 Min[1] := -3000;
 Max[1] := +3000;
 NewtonSetWorldSize(pNewtonWorld, @Min[0], @Max[0]);
 end;
end;

// =============================================================================
//   TTerrain.Render
// =============================================================================
procedure TTerrain.Render(pCameraPosition : TVector3f; pMaxDrawDistance : Single);
begin
if glIsTexture(DetailTexture) then
 begin
 glActiveTexture(GL_TEXTURE1);
 glEnable(GL_TEXTURE_2D);
 glBindTexture(GL_TEXTURE_2D, DetailTexture);
 end;
glActiveTexture(GL_TEXTURE0);
glEnable(GL_TEXTURE_2D);
glBindTexture(GL_TEXTURE_2D, BaseTexture);
OctreeSettings.CameraPos := glVertex(pCameraPosition[0], pCameraPosition[1], pCameraPosition[2]);
OctreeStats.TotalNodesDrawn := 0;
OctreeStats.TotalTrisDrawn  := 0;
OctreeSettings.MaxDrawDistance := pMaxDrawDistance;
glEnable(GL_CULL_FACE);
glCullFace(GL_FRONT);
Octree.DrawOctree(Octree);
glCullFace(GL_BACK);
glActiveTexture(GL_TEXTURE1);
glDisable(GL_TEXTURE_2D);
glActiveTexture(GL_TEXTURE0);
end;

// =============================================================================
//   TTerrain.Destroy
// =============================================================================
destructor TTerrain.Destroy;
begin
NewtonDestroyBody(NewtonWorld, NewtonBody);
if Assigned(Octree) then
 Octree.Free;
glDeleteTextures(1, @BaseTexture);
glDeleteTextures(1, @DetailTexture);
end;

end.
