// =============================================================================
//
//  glOctree.pas
//
// =============================================================================
//
//   Copyright  2005 by Sascha Willems (www.delphigl.de)
//   Newton Game Dynamics  2003/04/05 by Julio Jerez (www.newtondynamics.com)
//
// =============================================================================
//
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
//
// =============================================================================
//
//   Entry for the 2005 PGD "Dog-Fight" Competition
//
// =============================================================================

unit glOctree;

interface

uses
 SysUtils,
 glFrustum,
 dglOpenGL;

type
 PglVertex3f = ^TglVertex3f;
 TglVertex3f = record
   x,y,z : Single;
   t,u   : Single;
  end;
 TVertexArray = array of TglVertex3f;

type
  TBoolArray = array of Boolean;

  TDebug = class
  private
    FLines: array of TglVertex3f;
  protected
  public
    procedure AddDebugLine(const Point1, Point2: TglVertex3f);
    procedure AddDebugRectangle(const Centre: TglVertex3f; Width, Height, Depth: Single);
    procedure Clear;
    procedure RenderDebugLines;
  end;

  TNodeCorners = (ncTopLeftFront, ncTopLeftBack, ncTopRightBack, ncTopRightFront, ncBottomLeftFront, ncBottomLeftBack,
                  ncBottomRightBack, ncBottomRightFront);

  TOctree = class
    FSubdivided    : Boolean;
    FWidth         : Single;
    FTriangleCount : Cardinal;
    FCentre        : TglVertex3f;
    FOctreeNodes   : array[TNodeCorners] of TOctree;
    FVertices      : array of TglVertex3f;
    FCurrentTexID  : glUInt;
    FListIndex     : Cardinal;
    property Centre: TglVertex3f read FCentre;
    property Subdivided: Boolean read FSubdivided;
    property TriangleCount: Cardinal read FTriangleCount;
    property Width: Single read FWidth;
    constructor Create;
    destructor Destroy; override;
    function GetNewNodeCentre(const vCentre: TglVertex3f; width_: Single; nodeID: TNodeCorners): TglVertex3f;
    procedure InitOctree;
    procedure DestroyOctree;
    procedure GetSceneDimensions(const pVertices: TVertexArray; numberOfVerts: Cardinal);
    procedure CreateNode(const pVertices: TVertexArray; numberOfVerts: Cardinal; const vCentre: TglVertex3f; width_: Single);
    procedure CreateNewNode(const pVertices: TVertexArray; var pList: TBoolArray; numberOfVerts: Cardinal; const vCentre: TglVertex3f; width: Single; triangleCount: Integer; nodeID: TNodeCorners);
    procedure AssignVerticesToNode(const pVertices: TVertexArray; numberOfVerts: Cardinal);
    procedure DrawOctree(pNode: TOctree; pNoVisibilityTest : Boolean = False);
    procedure ShowDebugLines;
  end;

var
 OctreeSettings : record
   RenderMode      : glUInt;
   MaxTriangles    : glUInt;
   MaxDrawDistance : Single;
   MaxSubdivisions : glUInt;
   VertexArray     : TVertexArray;
   CameraPos       : TglVertex3f;
  end;
 OctreeStats : record
   EndNodeCount       : Cardinal;
   CurrentSubdivision : Cardinal;
   NumOfVerts         : Cardinal;
   TotalNodesDrawn    : Cardinal;
   TotalTrisDrawn     : Cardinal;
  end;

var
 g_Debug : TDebug = nil;

function glVertex(pX, pY, pZ : Single) : TglVertex3f;

implementation

function glVertex(pX, pY, pZ : Single) : TglVertex3f;
begin
Result.x := pX;
Result.y := pY;
Result.z := pZ;
end;

function glAddVector(pV1, pV2 : TglVertex3f) : TglVertex3f;
begin
Result.x := pV1.x + pV2.x;
Result.y := pV1.y + pV2.y;
Result.z := pV1.z + pV2.z;
end;

function glVectorDistance(v1, v2 : TglVertex3f) : Single;
begin
Result := Sqrt(Sqr(v1.x-v2.x)+Sqr(v1.y-v2.y)+Sqr(v1.z-v2.z));
end;

constructor TOctree.Create;
begin
inherited;
InitOctree;
end;

destructor TOctree.Destroy;
begin
DestroyOctree;
inherited;
end;

procedure TOctree.InitOctree;
var
  i: TNodeCorners;
begin
FSubdivided := False;
FWidth := 0;
FTriangleCount := 0;
SetLength(FVertices, 0);
for i := Low(FOctreeNodes) to High(FOctreeNodes) do
 FOctreeNodes[i] := nil;
end;

procedure TOctree.DestroyOctree;
var
 i : TNodeCorners;
begin
SetLength(FVertices, 0);
for i := Low(FOctreeNodes) to High(FOctreeNodes) do
 FreeAndNil(FOctreeNodes[i]);
g_Debug.Clear;
InitOctree;
end;

procedure TOctree.GetSceneDimensions(const pVertices: TVertexArray; numberOfVerts: Cardinal);
var
  currentWidth, currentHeight, currentDepth: Single;
  maxWidth, maxHeight, maxDepth: Single;
  i: Cardinal;
begin
maxWidth := 0;
maxHeight := 0;
maxDepth := 0;
if Length(pVertices) = 0 then Exit;
FCentre := glVertex(0,0,0);
for i := 0 to numberOfVerts - 1 do
 FCentre := glAddVector(FCentre, pVertices[i]);
FCentre.x := FCentre.x / numberOfVerts;
FCentre.y := FCentre.y / numberOfVerts;
FCentre.z := FCentre.z / numberOfVerts;
for i := 0 to numberOfVerts - 1 do
 begin
 currentWidth  := abs(pVertices[i].x - FCentre.x);
 currentHeight := abs(pVertices[i].y - FCentre.y);
 currentDepth  := abs(pVertices[i].z - FCentre.z);
 if currentWidth  > maxWidth then maxWidth := currentWidth;
 if currentHeight > maxHeight then maxHeight := currentHeight;
 if currentDepth > maxDepth then maxDepth := currentDepth;
 end;
maxWidth  := maxWidth  * 2;
maxHeight := maxHeight * 2;
maxDepth  := maxDepth  * 2;
if (maxWidth > maxHeight) and (maxWidth > maxDepth) then
 FWidth := maxWidth
else
 if(maxHeight > maxWidth) and (maxHeight > maxDepth) then
 	FWidth := maxHeight
 else
	FWidth := maxDepth;
end;

function TOctree.GetNewNodeCentre(const vCentre: TglVertex3f; width_: Single; nodeID: TNodeCorners): TglVertex3f;
begin
case NodeID of
 ncTopLeftFront     : Result := glVertex(vCentre.x - width/4, vCentre.y + width/4, vCentre.z + width/4);
 ncTopLeftBack      : Result := glVertex(vCentre.x - width/4, vCentre.y + width/4, vCentre.z - width/4);
 ncTopRightBack     : Result := glVertex(vCentre.x + width/4, vCentre.y + width/4, vCentre.z - width/4);
 ncTopRightFront    : Result := glVertex(vCentre.x + width/4, vCentre.y + width/4, vCentre.z + width/4);
 ncBottomLeftFront  : Result := glVertex(vCentre.x - width/4, vCentre.y - width/4, vCentre.z + width/4);
 ncBottomLeftBack   : Result := glVertex(vCentre.x - width/4, vCentre.y - width/4, vCentre.z - width/4);
 ncBottomRightBack  : Result := glVertex(vCentre.x + width/4, vCentre.y - width/4, vCentre.z - width/4);
 ncBottomRightFront : Result := glVertex(vCentre.x + width/4, vCentre.y - width/4, vCentre.z + width/4);
 end;
end;

procedure TOctree.CreateNode(const pVertices: TVertexArray; numberOfVerts: Cardinal; const vCentre: TglVertex3f; width_: Single);
var
  numberOfTriangles: Cardinal;
  i: Cardinal;
  triCount1,triCount2,triCount3,triCount4,triCount5,triCount6,triCount7,triCount8: Cardinal;
  pList1,pList2,pList3,pList4,pList5,pList6,pList7,pList8: TBoolArray;
begin
numberOfTriangles := numberOfVerts div 3;
FCentre := vCentre;
FWidth := width_;
g_Debug.AddDebugRectangle(Centre, width, width, width);
if((numberOfTriangles > OctreeSettings.MaxTriangles) and
   (OctreeStats.CurrentSubdivision < OctreeSettings.MaxSubdivisions) ) then
 	 begin
   FSubDivided := true;
	 SetLength(pList1, numberOfTriangles);
	 SetLength(pList2, numberOfTriangles);
	 SetLength(pList3, numberOfTriangles);
	 SetLength(pList4, numberOfTriangles);
	 SetLength(pList5, numberOfTriangles);
	 SetLength(pList6, numberOfTriangles);
	 SetLength(pList7, numberOfTriangles);
	 SetLength(pList8, numberOfTriangles);
   for i := 0 to numberOfVerts - 1 do
    begin
	  pList1[i div 3] := ((pVertices[i].x <= vCentre.x) and (pVertices[i].y >= vCentre.y) and (pVertices[i].z >= vCentre.z));
	  pList2[i div 3] := ((pVertices[i].x <= vCentre.x) and (pVertices[i].y >= vCentre.y) and (pVertices[i].z <= vCentre.z));
		pList3[i div 3] := ((pVertices[i].x >= vCentre.x) and (pVertices[i].y >= vCentre.y) and (pVertices[i].z <= vCentre.z));
		pList4[i div 3] := ((pVertices[i].x >= vCentre.x) and (pVertices[i].y >= vCentre.y) and (pVertices[i].z >= vCentre.z));
		pList5[i div 3] := ((pVertices[i].x <= vCentre.x) and (pVertices[i].y <= vCentre.y) and (pVertices[i].z >= vCentre.z));
		pList6[i div 3] := ((pVertices[i].x <= vCentre.x) and (pVertices[i].y <= vCentre.y) and (pVertices[i].z <= vCentre.z));
		pList7[i div 3] := ((pVertices[i].x >= vCentre.x) and (pVertices[i].y <= vCentre.y) and (pVertices[i].z <= vCentre.z));
		pList8[i div 3] := ((pVertices[i].x >= vCentre.x) and (pVertices[i].y <= vCentre.y) and (pVertices[i].z >= vCentre.z));
 		end;
	triCount1 := 0;
  triCount2 := 0;
  triCount3 := 0;
  triCount4 := 0;
  triCount5 := 0;
  triCount6 := 0;
  triCount7 := 0;
  triCount8 := 0;
  for i := 0 to numberOfTriangles - 1 do
   begin
	 if pList1[i] then Inc(triCount1);
   if pList2[i] then Inc(triCount2);
	 if pList3[i] then Inc(triCount3);
   if pList4[i] then Inc(triCount4);
	 if pList5[i] then Inc(triCount5);
   if pList6[i] then Inc(triCount6);
	 if pList7[i] then Inc(triCount7);
   if pList8[i] then Inc(triCount8);
	 end;
	CreateNewNode(pVertices, pList1, numberOfVerts, vCentre, width, triCount1, ncTopLeftFront);
  CreateNewNode(pVertices, pList2, numberOfVerts, vCentre, width, triCount2, ncTopLeftBack);
  CreateNewNode(pVertices, pList3, numberOfVerts, vCentre, width, triCount3, ncTopRightBack);
	CreateNewNode(pVertices, pList4, numberOfVerts, vCentre, width, triCount4, ncTopRightFront);
	CreateNewNode(pVertices, pList5, numberOfVerts, vCentre, width, triCount5, ncBottomLeftFront);
	CreateNewNode(pVertices, pList6, numberOfVerts, vCentre, width, triCount6, ncBottomLeftBack);
	CreateNewNode(pVertices, pList7, numberOfVerts, vCentre, width, triCount7, ncBottomRightBack);
	CreateNewNode(pVertices, pList8, numberOfVerts, vCentre, width, triCount8, ncBottomRightFront);
	end
 else
	AssignVerticesToNode(pVertices, numberOfVerts);
end;

procedure TOctree.CreateNewNode(const pVertices: TVertexArray; var pList: TBoolArray; numberOfVerts: Cardinal; const vCentre: TglVertex3f; width: Single; triangleCount: Integer; nodeID: TNodeCorners);
var
  pNodeVertices: TVertexArray;
  index: Cardinal;
  i: Cardinal;
  vNodeCentre: TglVertex3f;
begin
if triangleCount > 0 then
 begin
 SetLength(pNodeVertices, triangleCount * 3);
 index := 0;
 for i := 0 to numberOfVerts - 1 do
  if pList[i div 3] then
	 begin
	 pNodeVertices[index] := pVertices[i];
	 Inc(index);
	 end;
 FOctreeNodes[nodeID] := TOctree.Create;
 vNodeCentre := GetNewNodeCentre(vCentre, width, nodeID);
 Inc(OctreeStats.CurrentSubdivision);
 FOctreeNodes[nodeID].CreateNode(pNodeVertices, triangleCount * 3, vNodeCentre, width / 2);
 Dec(OctreeStats.CurrentSubdivision);
 SetLength(pNodeVertices, 0);
 end;
end;

procedure TOctree.AssignVerticesToNode(const pVertices: TVertexArray; numberOfVerts: Cardinal);
begin
FSubDivided := False;
FTriangleCount := numberOfVerts div 3;
SetLength(FVertices, numberOfVerts);
FillChar(FVertices[0], SizeOf(TglVertex3f) * numberOfVerts, 0);
Move(pVertices[0], FVertices[0], SizeOf(TglVertex3f) * numberOfVerts);
Inc(OctreeStats.EndNodeCount);
end;

procedure TOctree.DrawOctree(pNode: TOctree; pNoVisibilityTest : Boolean = False);
type
 TVectorArray = array[0..0] of TglVertex3f;
var
 i,v      : Cardinal;
 Max      : Cardinal;
 Distance : Single;
 V1       : TglVertex3f;
begin
if pNode = nil then
 exit;
if not pNoVisibilityTest then
 if not Frustum.IsBoxWithin(pNode.Centre.x, pNode.Centre.y, pNode.Centre.z, pNode.Width/2, pNode.Width/2, pNode.Width/2) then
  exit;
V1       := glVertex(pNode.Centre.x, OctreeSettings.CameraPos.y, pNode.Centre.z);
Distance := Abs(glVectorDistance(V1, OctreeSettings.CameraPos));
if not pNode.Subdivided then
 if Distance > OctreeSettings.MaxDrawDistance then
  exit;
if pNode.SubDivided then
 begin
 DrawOctree(pNode.FOctreeNodes[ncTopLeftFront], pNoVisibilityTest);
 DrawOctree(pNode.FOctreeNodes[ncTopLeftBack], pNoVisibilityTest);
 DrawOctree(pNode.FOctreeNodes[ncTopRightBack], pNoVisibilityTest);
 DrawOctree(pNode.FOctreeNodes[ncTopRightFront], pNoVisibilityTest);
 DrawOctree(pNode.FOctreeNodes[ncBottomLeftFront], pNoVisibilityTest);
 DrawOctree(pNode.FOctreeNodes[ncBottomLeftBack], pNoVisibilityTest);
 DrawOctree(pNode.FOctreeNodes[ncBottomRightBack], pNoVisibilityTest);
 DrawOctree(pNode.FOctreeNodes[ncBottomRightFront], pNoVisibilityTest);
 end
else
 begin
 if glIsList(pNode.FListIndex) then
  begin
  Inc(OctreeStats.TotalNodesDrawn);
  Inc(OctreeStats.TotalTrisDrawn, pNode.TriangleCount);
  // Fade if this node is nearing the max visiblity distance
  if Distance > OctreeSettings.MaxDrawDistance-384 then
   begin
   glEnable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glColor4f(1,1,1, (OctreeSettings.MaxDrawDistance-Distance)/384);
   end;
  glCallList(pNode.FListIndex);
  glDisable(GL_BLEND);
  glColor3f(1,1,1);
  end
 else
  begin
  if pNode.FVertices = nil then
   exit;
  FCurrentTexID := 0;
  pNode.FListIndex := glGenLists(1);
  glNewList(pNode.FListIndex, GL_COMPILE);
   i   := 0;
   Max := pNode.TriangleCount * 3;
   glBegin(OctreeSettings.RenderMode);
    while i < Max do
     begin
     for v := 0 to 2 do
      begin
      glMultiTexCoord2f(GL_TEXTURE0, pNode.FVertices[i+v].t, pNode.FVertices[i+v].u);
      glMultiTexCoord2f(GL_TEXTURE1, pNode.FVertices[i+v].t*32, pNode.FVertices[i+v].u*32);
      glVertex3fv(@pNode.FVertices[i+v]);
      end;
     inc(i, 3);
     end;
   glEnd;
  glEndList;
  end;
 end;
end;

procedure TDebug.AddDebugLine(const Point1, Point2: TglVertex3f);
var
  OldSize: Integer;
begin
  OldSize := Length(FLines);
  SetLength(FLines, OldSize + 2);
  FLines[OldSize] := Point1;
  FLines[OldSize + 1] := Point2;
end;

procedure TDebug.AddDebugRectangle(const Centre: TglVertex3f; Width, Height, Depth: Single);
var
  vTopLeftFront: TglVertex3f;
  vTopLeftBack : TglVertex3f;
  vTopRightBack: TglVertex3f;
  vTopRightFront: TglVertex3f;
  vBottom_LeftFront: TglVertex3f;
  vBottom_LeftBack: TglVertex3f;
  vBottomRightBack: TglVertex3f;
  vBottomRightFront: TglVertex3f;
  OldSize: Integer;
begin
width  := width / 2;
height := height / 2;
depth  := depth / 2;
vTopLeftFront  := glVertex(Centre.x - width, Centre.y + height,  Centre.z + depth);
vTopLeftBack   := glVertex(Centre.x - width, Centre.y + height,  Centre.z - depth);
vTopRightBack  := glVertex(Centre.x + width, Centre.y + height,  Centre.z - depth);
vTopRightFront := glVertex(Centre.x + width, Centre.y + height, Centre.z + depth);
vBottom_LeftFront := glVertex(Centre.x - width, Centre.y - height, Centre.z + depth);
vBottom_LeftBack  := glVertex(Centre.x - width, Centre.y - height, Centre.z - depth);
vBottomRightBack  := glVertex(Centre.x + width, Centre.y - height, Centre.z - depth);
vBottomRightFront := glVertex(Centre.x + width, Centre.y - height, Centre.z + depth);
OldSize := Length(FLines);
SetLength(FLines, OldSize + 24);
FLines[OldSize]      := vTopLeftFront;		 FLines[OldSize + 1] := vTopRightFront;
FLines[OldSize + 2]  := vTopLeftBack;  		 FLines[OldSize + 3] := vTopRightBack;
FLines[OldSize + 4]  := vTopLeftFront;		 FLines[OldSize + 5] := vTopLeftBack;
FLines[OldSize + 6]  := vTopRightFront;	   FLines[OldSize + 7] := vTopRightBack;
FLines[OldSize + 8]  := vBottom_LeftFront; FLines[OldSize + 9] := vBottomRightFront;
FLines[OldSize + 10] := vBottom_LeftBack;	 FLines[OldSize + 11] := vBottomRightBack;
FLines[OldSize + 12] := vBottom_LeftFront; FLines[OldSize + 13] := vBottom_LeftBack;
FLines[OldSize + 14] := vBottomRightFront; FLines[OldSize + 15] := vBottomRightBack;
FLines[OldSize + 16] := vTopLeftFront;		 FLines[OldSize + 17] := vBottom_LeftFront;
FLines[OldSize + 18] := vTopLeftBack;		   FLines[OldSize + 19] := vBottom_LeftBack;
FLines[OldSize + 20] := vTopRightBack;		 FLines[OldSize + 21] := vBottomRightBack;
FLines[OldSize + 22] := vTopRightFront;		 FLines[OldSize + 23] := vBottomRightFront;
end;

procedure TDebug.Clear;
begin
SetLength(FLines, 0);
end;

procedure TDebug.RenderDebugLines;
var
 i : Integer;
begin
glBegin(GL_LINES);
 for i := 0 to Length(FLines) - 1 do
 	glVertex3fv(@FLines[i]);
glEnd;
end;

procedure TOctree.ShowDebugLines;
begin
g_Debug.RenderDebugLines;
end;

initialization
  g_Debug := TDebug.Create;
  with OctreeSettings do
   begin
   RenderMode      := GL_TRIANGLES;
   MaxTriangles    := 1024;
   MaxDrawDistance := 2048;
   MaxSubdivisions := 4;
   end;

finalization
  g_Debug.Free;

end.
